"""The MySQL backend."""

from __future__ import annotations

import contextlib
import warnings
from functools import cached_property
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import MySQLdb
import sqlglot as sg
import sqlglot.expressions as sge
from MySQLdb import ProgrammingError
from MySQLdb.constants import ER

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateDatabase
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import STAR, TRUE, C, RenameTable

if TYPE_CHECKING:
    from collections.abc import Mapping
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa


class Backend(SQLBackend, CanCreateDatabase):
    name = "mysql"
    compiler = sc.mysql.compiler
    supports_create_or_replace = False

    def _from_url(self, url: ParseResult, **kwargs):
        """Connect to a backend using a URL `url`.

        Parameters
        ----------
        url
            URL with which to connect to a backend.
        kwargs
            Additional keyword arguments

        Returns
        -------
        BaseBackend
            A backend instance

        """
        database, *_ = url.path[1:].split("/", 1)
        connect_args = {
            "user": url.username,
            "password": unquote_plus(url.password or ""),
            "host": url.hostname,
            "database": database or "",
            "port": url.port or None,
        }

        kwargs.update(connect_args)
        self._convert_kwargs(kwargs)

        if "user" in kwargs and not kwargs["user"]:
            del kwargs["user"]

        if "host" in kwargs and not kwargs["host"]:
            del kwargs["host"]

        if "database" in kwargs and not kwargs["database"]:
            del kwargs["database"]

        if "password" in kwargs and kwargs["password"] is None:
            del kwargs["password"]

        if "port" in kwargs and kwargs["port"] is None:
            del kwargs["port"]

        return self.connect(**kwargs)

    @cached_property
    def version(self):
        return ".".join(map(str, self.con._server_version))

    def do_connect(
        self,
        host: str = "localhost",
        user: str | None = None,
        password: str | None = None,
        port: int = 3306,
        autocommit: bool = True,
        **kwargs,
    ) -> None:
        """Create an Ibis client using the passed connection parameters.

        Parameters
        ----------
        host
            Hostname
        user
            Username
        password
            Password
        port
            Port
        autocommit
            Autocommit mode
        kwargs
            Additional keyword arguments passed to `MySQLdb.connect`

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_MYSQL_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_MYSQL_USER", "ibis")
        >>> password = os.environ.get("IBIS_TEST_MYSQL_PASSWORD", "ibis")
        >>> database = os.environ.get("IBIS_TEST_MYSQL_DATABASE", "ibis-testing")
        >>> con = ibis.mysql.connect(database=database, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int32
          bool_col        int8
          tinyint_col     int8
          smallint_col    int16
          int_col         int32
          bigint_col      int64
          float_col       float32
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp
          year            int32
          month           int32
        """
        self.con = MySQLdb.connect(
            user=user,
            host="127.0.0.1" if host == "localhost" else host,
            port=port,
            password=password,
            autocommit=autocommit,
            **kwargs,
        )

        self._post_connect()

    @util.experimental
    @classmethod
    def from_connection(cls, con) -> Backend:
        """Create an Ibis client from an existing connection to a MySQL database.

        Parameters
        ----------
        con
            An existing connection to a MySQL database.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect()
        return new_backend

    def _post_connect(self) -> None:
        with self.con.cursor() as cur:
            try:
                cur.execute("SET @@session.time_zone = 'UTC'")
            except Exception as e:  # noqa: BLE001
                warnings.warn(f"Unable to set session timezone to UTC: {e}")

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.database())) as cur:
            [(database,)] = cur.fetchall()
        return database

    def list_databases(self, like: str | None = None) -> list[str]:
        # In MySQL, "database" and "schema" are synonymous
        with self._safe_raw_sql("SHOW DATABASES") as cur:
            databases = list(map(itemgetter(0), cur.fetchall()))
        return self._filter_with_like(databases, like)

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        from ibis.backends.mysql.datatypes import _type_from_cursor_info

        sql = (
            sg.select(STAR)
            .from_(
                sg.parse_one(query, dialect=self.dialect).subquery(
                    sg.to_identifier("tmp", quoted=self.compiler.quoted)
                )
            )
            .limit(0)
            .sql(self.dialect)
        )
        with self.begin() as cur:
            cur.execute(sql)
            descr, flags = cur.description, cur.description_flags

        items = {}
        for (name, type_code, _, _, field_length, scale, _), raw_flags in zip(
            descr, flags
        ):
            item = _type_from_cursor_info(
                flags=raw_flags,
                type_code=type_code,
                field_length=field_length,
                scale=scale,
            )
            items[name] = item
        return sch.Schema(items)

    def get_schema(
        self, name: str, *, catalog: str | None = None, database: str | None = None
    ) -> sch.Schema:
        table = sg.table(
            name, db=database, catalog=catalog, quoted=self.compiler.quoted
        ).sql(self.dialect)

        with self.begin() as cur:
            try:
                cur.execute(sge.Describe(this=table).sql(self.dialect))
            except ProgrammingError as e:
                if e.args[0] == ER.NO_SUCH_TABLE:
                    raise com.TableNotFound(name) from e
            else:
                result = cur.fetchall()

        type_mapper = self.compiler.type_mapper
        fields = {
            name: type_mapper.from_string(type_string, nullable=is_nullable == "YES")
            for name, type_string, is_nullable, *_ in result
        }

        return sch.Schema(fields)

    def create_database(self, name: str, force: bool = False) -> None:
        sql = sge.Create(kind="DATABASE", exist=force, this=sg.to_identifier(name)).sql(
            self.name
        )
        with self.begin() as cur:
            cur.execute(sql)

    def drop_database(self, name: str, force: bool = False) -> None:
        sql = sge.Drop(kind="DATABASE", exist=force, this=sg.to_identifier(name)).sql(
            self.name
        )
        with self.begin() as cur:
            cur.execute(sql)

    @contextlib.contextmanager
    def begin(self):
        con = self.con
        cur = con.cursor()
        autocommit = con.get_autocommit()

        if not autocommit:
            con.begin()

        try:
            yield cur
        except Exception:
            if not autocommit:
                con.rollback()
            raise
        else:
            if not autocommit:
                con.commit()
        finally:
            cur.close()

    # TODO(kszucs): should make it an abstract method or remove the use of it
    # from .execute()
    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        with self.raw_sql(*args, **kwargs) as result:
            yield result

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.name)

        con = self.con
        autocommit = con.get_autocommit()

        cursor = con.cursor()

        if not autocommit:
            con.begin()

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            if not autocommit:
                con.rollback()
            cursor.close()
            raise
        else:
            if not autocommit:
                con.commit()
            return cursor

    # TODO: disable positional arguments
    def list_tables(
        self,
        like: str | None = None,
        database: tuple[str, str] | str | None = None,
    ) -> list[str]:
        """List the tables in the database.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of tables is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Database to list tables from. Default behavior is to show tables in
            the current database (`self.current_database`).
        """
        if database is not None:
            table_loc = self._to_sqlglot_table(database)
        else:
            table_loc = sge.Table(
                db=sg.to_identifier(self.current_database, quoted=self.compiler.quoted),
                catalog=None,
            )

        conditions = [TRUE]

        if (sg_cat := table_loc.args["catalog"]) is not None:
            sg_cat.args["quoted"] = False
        if (sg_db := table_loc.args["db"]) is not None:
            sg_db.args["quoted"] = False
        if table_loc.catalog or table_loc.db:
            conditions = [C.table_schema.eq(sge.convert(table_loc.sql(self.name)))]

        col = "table_name"
        sql = (
            sg.select(col)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
            .where(*conditions)
            .sql(self.name)
        )

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def execute(
        self, expr: ir.Expr, limit: str | None = "default", **kwargs: Any
    ) -> Any:
        """Execute an expression."""

        self._run_pre_execute_hooks(expr)
        table = expr.as_table()
        sql = self.compile(table, limit=limit, **kwargs)

        schema = table.schema()

        with self._safe_raw_sql(sql) as cur:
            result = self._fetch_from_cursor(cur, schema)
        return expr.__pandas_result__(result)

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        if not schema:
            schema = table.schema()

        quoted = self.compiler.quoted
        dialect = self.dialect

        table_expr = sg.table(temp_name, catalog=database, quoted=quoted)
        target = sge.Schema(this=table_expr, expressions=schema.to_sqlglot(dialect))

        create_stmt = sge.Create(
            kind="TABLE", this=target, properties=sge.Properties(expressions=properties)
        )

        this = sg.table(name, catalog=database, quoted=quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                cur.execute(sge.Insert(this=table_expr, expression=query).sql(dialect))

            if overwrite:
                cur.execute(sge.Drop(kind="TABLE", this=this, exists=True).sql(dialect))
                cur.execute(
                    sge.Alter(
                        kind="TABLE",
                        this=table_expr,
                        exists=True,
                        actions=[RenameTable(this=this)],
                    ).sql(dialect)
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "MySQL cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        name = op.name
        quoted = self.compiler.quoted
        dialect = self.dialect

        create_stmt = sg.exp.Create(
            kind="TABLE",
            this=sg.exp.Schema(
                this=sg.to_identifier(name, quoted=quoted),
                expressions=schema.to_sqlglot(dialect),
            ),
            properties=sg.exp.Properties(expressions=[sge.TemporaryProperty()]),
        )
        create_stmt_sql = create_stmt.sql(dialect)

        df = op.data.to_frame()
        # nan can not be used with MySQL
        df = df.replace(float("nan"), None)

        data = df.itertuples(index=False)
        sql = self._build_insert_template(
            name, schema=schema, columns=True, placeholder="%s"
        )
        with self.begin() as cur:
            cur.execute(create_stmt_sql)

            if not df.empty:
                cur.executemany(sql, data)

    @util.experimental
    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1_000_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        import pyarrow as pa

        self._run_pre_execute_hooks(expr)

        schema = expr.as_table().schema()
        with self._safe_raw_sql(
            self.compile(expr, limit=limit, params=params)
        ) as cursor:
            df = self._fetch_from_cursor(cursor, schema)
        table = pa.Table.from_pandas(
            df, schema=schema.to_pyarrow(), preserve_index=False
        )
        return table.to_reader(max_chunksize=chunk_size)

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        import pandas as pd

        from ibis.backends.mysql.converter import MySQLPandasData

        df = pd.DataFrame.from_records(
            cursor.fetchall(), columns=schema.names, coerce_float=True
        )
        return MySQLPandasData.convert_table(df, schema)



from __future__ import annotations

import math
import statistics
from collections import Counter
from datetime import datetime
from functools import partial

import pytest
import pytz
import toolz
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis.backends.tests.errors import (
    ClickHouseDatabaseError,
    DatabricksServerOperationError,
    GoogleBadRequest,
    MySQLOperationalError,
    PolarsComputeError,
    PsycoPg2ArraySubscriptError,
    PsycoPg2IndeterminateDatatype,
    PsycoPg2InternalError,
    PsycoPg2ProgrammingError,
    PsycoPg2SyntaxError,
    Py4JJavaError,
    PySparkAnalysisException,
    TrinoUserError,
)
from ibis.common.collections import frozendict
from ibis.conftest import IS_SPARK_REMOTE

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")
tm = pytest.importorskip("pandas.testing")

pytestmark = [
    pytest.mark.never(
        ["sqlite", "mysql", "exasol"], reason="No array support", raises=Exception
    ),
    pytest.mark.never(
        ["mssql"],
        reason="No array support",
        raises=(
            com.UnsupportedBackendType,
            com.OperationNotDefinedError,
            AssertionError,
        ),
    ),
    pytest.mark.never(
        ["mysql"],
        reason="No array support",
        raises=(
            com.UnsupportedBackendType,
            com.OperationNotDefinedError,
            MySQLOperationalError,
        ),
    ),
    pytest.mark.notyet(
        ["impala"],
        reason="No array support",
        raises=(
            com.UnsupportedBackendType,
            com.OperationNotDefinedError,
            com.TableNotFound,
        ),
    ),
    pytest.mark.notimpl(["druid", "oracle"], raises=Exception),
]

# NB: We don't check whether results are numpy arrays or lists because this
# varies across backends. At some point we should unify the result type to be
# list.


def test_array_column(backend, alltypes, df):
    expr = ibis.array(
        [alltypes["double_col"], alltypes["double_col"], 5.0, ibis.literal(6.0)]
    )
    assert isinstance(expr, ir.ArrayColumn)

    result = expr.execute()
    expected = df.apply(
        lambda row: [row["double_col"], row["double_col"], 5.0, 6.0],
        axis=1,
    )
    backend.assert_series_equal(result, expected, check_names=False)


ARRAY_BACKEND_TYPES = {
    "clickhouse": "Array(Float64)",
    "snowflake": "ARRAY",
    "trino": "array(double)",
    "bigquery": "ARRAY",
    "duckdb": "DOUBLE[]",
    "postgres": "numeric[]",
    "risingwave": "numeric[]",
    "flink": "ARRAY<DECIMAL(2, 1) NOT NULL> NOT NULL",
}


def test_array_scalar(con):
    expr = ibis.array([1.0, 2.0, 3.0])
    assert isinstance(expr, ir.ArrayScalar)

    result = con.execute(expr.name("tmp"))
    expected = np.array([1.0, 2.0, 3.0])

    assert np.array_equal(result, expected)


@pytest.mark.notimpl(["flink", "polars"], raises=com.OperationNotDefinedError)
def test_array_repeat(con):
    expr = ibis.array([1.0, 2.0]) * 2

    result = con.execute(expr.name("tmp"))
    expected = np.array([1.0, 2.0, 1.0, 2.0])

    assert np.array_equal(result, expected)


@pytest.mark.notimpl(["flink", "polars"], raises=com.OperationNotDefinedError)
def test_array_repeat_column(con):
    t = ibis.memtable({"x": [[1.0, 2.0]]}, schema=ibis.schema({"x": "array<float64>"}))
    expr = (t.x * 2).name("tmp")

    result = con.execute(expr.name("tmp")).iat[0]
    expected = np.array([1.0, 2.0, 1.0, 2.0])
    assert Counter(result) == Counter(expected)


def test_array_concat(con):
    left = ibis.literal([1, 2, 3])
    right = ibis.literal([2, 1])
    expr = left + right
    result = con.execute(expr.name("tmp"))
    assert sorted(result) == sorted([1, 2, 3, 2, 1])


def test_array_concat_variadic(con):
    left = ibis.literal([1, 2, 3])
    right = ibis.literal([2, 1])
    expr = left.concat(right, right, right)
    result = con.execute(expr.name("tmp"))
    expected = np.array([1, 2, 3, 2, 1, 2, 1, 2, 1])
    assert np.array_equal(result, expected)


# Issues #2370
@pytest.mark.notimpl(["flink"], raises=Py4JJavaError)
@pytest.mark.notyet(["trino"], raises=TrinoUserError)
def test_array_concat_some_empty(con):
    left = ibis.literal([])
    right = ibis.literal([2, 1])
    expr = left.concat(right)
    result = con.execute(expr.name("tmp"))
    expected = np.array([2, 1])
    assert np.array_equal(result, expected)


def test_array_radd_concat(con):
    left = [1]
    right = ibis.literal([2])
    expr = left + right
    result = con.execute(expr.name("tmp"))
    expected = np.array([1, 2])

    assert np.array_equal(result, expected)


@pytest.mark.parametrize("op", [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat_scalar(con, op):
    raw_left = [1, 2, 3]
    raw_right = [3, 4]
    left = ibis.literal(raw_left)
    right = ibis.literal(raw_right)
    expr = op(left, right)
    result = con.execute(expr)
    assert result == op(raw_left, raw_right)


@pytest.mark.parametrize(
    "op", [lambda x, y: x + y, lambda x, y: y + x], ids=["left", "right"]
)
@pytest.mark.notyet(
    ["clickhouse"],
    raises=ClickHouseDatabaseError,
    reason="doesn't support nullable arrays",
)
@pytest.mark.notyet(
    ["bigquery"],
    raises=ValueError,
    reason=(
        "bigquery treats null arrays as empty "
        "and a ValueError is raised from `pd.isna` "
        "when the value is not a truthy value"
    ),
)
def test_array_concat_with_null(con, op):
    non_null_value = ibis.array([2**31 - 1])
    null_value = ibis.null(non_null_value.type())
    expr = op(non_null_value, null_value)
    result = con.execute(expr)
    assert pd.isna(result)


@pytest.mark.notyet(
    ["clickhouse", "bigquery"],
    raises=AssertionError,
    reason="doesn't support nullable arrays, so result is not null",
)
@pytest.mark.notyet(
    ["datafusion"],
    raises=Exception,
    reason="cannot handle concatenating two null values that are typed as arrays",
)
def test_array_concat_with_null_non_constant(con):
    t = ibis.memtable({"a": [None]}, schema={"a": "array<int32>"})
    expr = t.a + t.a
    result = con.execute(expr)
    assert all(pd.isna(result))


def test_array_length(con):
    expr = ibis.literal([1, 2, 3]).length()
    assert con.execute(expr.name("tmp")) == 3


def test_list_literal(con):
    arr = [1, 2, 3]
    expr = ibis.literal(arr)
    result = con.execute(expr.name("tmp"))

    assert np.array_equal(result, arr)


def test_np_array_literal(con):
    arr = np.array([1, 2, 3])
    expr = ibis.literal(arr)
    result = con.execute(expr.name("tmp"))

    assert np.array_equal(result, arr)


@pytest.mark.parametrize("idx", range(3))
def test_array_index(con, idx):
    arr = [1, 2, 3]
    expr = ibis.literal(arr)
    expr = expr[idx]
    result = con.execute(expr)
    assert result == arr[idx]


builtin_array = toolz.compose(
    # these will almost certainly never be supported
    pytest.mark.never(
        ["mysql"],
        reason="array types are unsupported",
        raises=(
            com.OperationNotDefinedError,
            MySQLOperationalError,
            com.UnsupportedBackendType,
        ),
    ),
    pytest.mark.never(
        ["sqlite"],
        reason="array types are unsupported",
        raises=(com.UnsupportedBackendType,),
    ),
)


@builtin_array
@pytest.mark.notyet(
    ["clickhouse", "postgres"],
    reason="backend does not support nullable nested types",
    raises=AssertionError,
)
@pytest.mark.notimpl(
    ["risingwave"],
    raises=AssertionError,
    reason="Do not nest ARRAY types; ARRAY(basetype) handles multi-dimensional arrays of basetype",
)
@pytest.mark.never(
    ["bigquery"], reason="doesn't support arrays of arrays", raises=AssertionError
)
@pytest.mark.never(
    ["snowflake"],
    reason="snowflake has an extremely specialized way of implementing arrays",
    raises=AssertionError,
)
def test_array_discovery(backend):
    t = backend.array_types
    expected = ibis.schema(
        dict(
            x=dt.Array(dt.int64),
            y=dt.Array(dt.string),
            z=dt.Array(dt.float64),
            grouper=dt.string,
            scalar_column=dt.float64,
            multi_dim=dt.Array(dt.Array(dt.int64)),
        )
    )
    assert t.schema() == expected


@builtin_array
@pytest.mark.notyet(
    ["bigquery"],
    reason="BigQuery doesn't support casting array<T> to array<U>",
    raises=GoogleBadRequest,
)
def test_unnest_simple(backend):
    array_types = backend.array_types
    expected = (
        array_types.execute()
        .x.explode()
        .reset_index(drop=True)
        .astype("Float64")
        .rename("tmp")
    )
    expr = array_types.x.cast("!array<float64>").unnest()
    result = expr.execute().astype("Float64").rename("tmp")
    assert frozenset(result.values) == frozenset(expected.values)


@builtin_array
def test_unnest_complex(backend):
    array_types = backend.array_types
    df = array_types.execute()
    expr = (
        array_types.select(["grouper", "x"])
        .mutate(x=lambda t: t.x.unnest())
        .group_by("grouper")
        .aggregate(count_flat=lambda t: t.x.count())
        .order_by("grouper")
    )
    expected = (
        df[["grouper", "x"]]
        .explode("x")
        .groupby("grouper")
        .x.count()
        .rename("count_flat")
        .reset_index()
        .sort_values("grouper")
        .reset_index(drop=True)
    )
    result = expr.execute()
    backend.assert_frame_equal(result, expected)

    # test that unnest works with to_pyarrow
    assert len(expr.to_pyarrow()) == len(result)


@builtin_array
def test_unnest_idempotent(backend):
    array_types = backend.array_types
    df = array_types.execute()
    expr = (
        array_types.select(
            ["scalar_column", array_types.x.cast("!array<int64>").unnest().name("x")]
        )
        .group_by("scalar_column")
        .aggregate(x=lambda t: t.x.collect())
        .mutate(x=lambda t: t.x.sort())
        .order_by("scalar_column")
    )
    result = expr.execute().reset_index(drop=True)
    expected = (
        df[["scalar_column", "x"]]
        .assign(x=df.x.map(lambda arr: sorted(i for i in arr if not pd.isna(i))))
        .sort_values("scalar_column")
        .reset_index(drop=True)
    )
    tm.assert_frame_equal(result, expected)


@builtin_array
def test_unnest_no_nulls(backend):
    array_types = backend.array_types
    df = array_types.execute()
    expr = (
        array_types.select(
            ["scalar_column", array_types.x.cast("!array<int64>").unnest().name("y")]
        )
        .filter(lambda t: t.y.notnull())
        .group_by("scalar_column")
        .aggregate(x=lambda t: t.y.collect())
        .order_by("scalar_column")
    )
    result = expr.execute().reset_index(drop=True)
    expected = (
        df[["scalar_column", "x"]]
        .explode("x")
        .dropna(subset=["x"])
        .groupby("scalar_column")
        .x.apply(lambda xs: [x for x in xs if x is not None])
        .reset_index()
    )
    tm.assert_frame_equal(result, expected)


@builtin_array
def test_unnest_default_name(backend):
    array_types = backend.array_types
    df = array_types.execute()
    expr = (
        array_types.x.cast("!array<int64>") + ibis.array([1]).cast("!array<int64>")
    ).unnest()
    assert expr.get_name().startswith("ArrayConcat(")

    result = expr.name("x").execute()
    expected = df.x.map(lambda x: x + [1]).explode("x")
    assert frozenset(result.astype(object).fillna(pd.NA).values) == frozenset(
        expected.fillna(pd.NA).values
    )


@pytest.mark.parametrize(
    ("start", "stop"),
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        (-3, None),
        (-3, -1),
        param(
            None,
            -3,
            marks=[
                pytest.mark.notyet(
                    ["flink"],
                    raises=AssertionError,
                    reason=(
                        "ArraySlice in Flink behaves unexpectedly when"
                        "`start` is None and `stop` is negative."
                    ),
                )
            ],
            id="nulls",
        ),
    ],
)
@pytest.mark.notimpl(["datafusion"], raises=com.OperationNotDefinedError)
def test_array_slice(backend, start, stop):
    array_types = backend.array_types
    expr = array_types.select(sliced=array_types.y[start:stop])
    result = expr.sliced.execute()
    expected = array_types.y.execute().map(lambda x: x[start:stop])
    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars", "sqlite"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="TODO(Kexiang): seems a bug",
)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array: ..."
)
@pytest.mark.parametrize(
    ("input", "output"),
    [
        param(
            {"a": [[1, None, 2], [4]]},
            {"a": [[2, None, 3], [5]]},
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="BigQuery doesn't support arrays with null elements",
                ),
                pytest.mark.notyet(
                    ["pyspark"],
                    condition=IS_SPARK_REMOTE,
                    raises=AssertionError,
                    reason="somehow, transformed results are different types",
                ),
                pytest.mark.notyet(
                    ["databricks"],
                    raises=AssertionError,
                    reason="nulls come back as NaN",
                ),
            ],
            id="nulls",
        ),
        param({"a": [[1, 2], [4]]}, {"a": [[2, 3], [5]]}, id="no_nulls"),
    ],
)
@pytest.mark.parametrize(
    "func",
    [lambda x: x + 1, partial(lambda x, y: x + y, y=1), ibis._ + 1],
    ids=["lambda", "partial", "deferred"],
)
def test_array_map(con, input, output, func):
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    expected = pd.Series(output["a"])

    expr = t.select(a=t.a.map(func))
    result = con.execute(expr.a)
    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars", "sqlite"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="TODO(Kexiang): seems a bug",
)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array: ..."
)
@pytest.mark.parametrize(
    ("input", "output"),
    [
        param(
            {"a": [[1, None, 2], [4]]},
            {"a": [[2, None, 5], [5]]},
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="BigQuery doesn't support arrays with null elements",
                ),
                pytest.mark.notyet(
                    ["pyspark"],
                    condition=IS_SPARK_REMOTE,
                    raises=AssertionError,
                    reason="somehow, transformed results are different types",
                ),
                pytest.mark.notimpl(
                    ["databricks"],
                    raises=AssertionError,
                    reason="nans instead of nulls",
                ),
            ],
            id="nulls",
        ),
        param({"a": [[1, 2], [4]]}, {"a": [[2, 4], [5]]}, id="no_nulls"),
    ],
)
@pytest.mark.parametrize(
    "func",
    [lambda x, i: x + 1 + i, partial(lambda x, y, i: x + y + i, y=1)],
    ids=["lambda", "partial"],
)
def test_array_map_with_index(con, input, output, func):
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    expected = pd.Series(output["a"])

    expr = t.select(a=t.a.map(func))
    result = con.execute(expr.a)
    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array..."
)
@pytest.mark.parametrize(
    ("input", "output"),
    [
        param(
            {"a": [[1, None, 2], [4]]},
            {"a": [[2], [4]]},
            id="nulls",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="NULLs are not allowed as array elements",
                )
            ],
        ),
        param({"a": [[1, 2], [4]]}, {"a": [[2], [4]]}, id="no_nulls"),
    ],
)
@pytest.mark.notyet(
    "risingwave",
    raises=PsycoPg2InternalError,
    reason="no support for not null column constraint",
)
@pytest.mark.parametrize(
    "predicate",
    [lambda x: x > 1, partial(lambda x, y: x > y, y=1), ibis._ > 1],
    ids=["lambda", "partial", "deferred"],
)
def test_array_filter(con, input, output, predicate):
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    expected = pd.Series(output["a"])

    expr = t.select(a=t.a.filter(predicate))
    result = con.execute(expr.a)
    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array..."
)
@pytest.mark.parametrize(
    ("input", "output"),
    [
        param(
            {"a": [[1, None, 2], [4]]},
            {"a": [[2], [4]]},
            id="nulls",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="NULLs are not allowed as array elements",
                )
            ],
        ),
        param({"a": [[1, 2], [4]]}, {"a": [[2], [4]]}, id="no_nulls"),
    ],
)
@pytest.mark.notyet(
    "risingwave",
    raises=PsycoPg2InternalError,
    reason="no support for not null column constraint",
)
@pytest.mark.parametrize(
    "predicate",
    [lambda x, i: x + (i - i) > 1, partial(lambda x, y, i: x > y + (i * 0), y=1)],
    ids=["lambda", "partial"],
)
def test_array_filter_with_index(con, input, output, predicate):
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))
    expected = pd.Series(output["a"])

    expr = t.select(a=t.a.filter(predicate))
    result = con.execute(expr.a)
    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array..."
)
@pytest.mark.parametrize(
    ("input", "output"),
    [
        param(
            {"a": [[1, None, None], [4]]},
            {"a": [[1, None], [4]]},
            id="nulls",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="NULLs are not allowed as array elements",
                )
            ],
        ),
        param({"a": [[1, 2], [1]]}, {"a": [[1], [1]]}, id="no_nulls"),
    ],
)
@pytest.mark.notyet(
    "risingwave",
    raises=PsycoPg2InternalError,
    reason="no support for not null column constraint",
)
@pytest.mark.parametrize(
    "predicate",
    [lambda x, i: i % 2 == 0, partial(lambda x, y, i: i % 2 == 0, y=1)],
    ids=["lambda", "partial"],
)
def test_array_filter_with_index_lambda(con, input, output, predicate):
    t = ibis.memtable(input, schema=ibis.schema(dict(a="!array<int8>")))

    expr = t.select(a=t.a.filter(predicate))
    result = con.to_pyarrow(expr.a)
    assert frozenset(map(tuple, result.to_pylist())) == frozenset(
        map(tuple, output["a"])
    )


@builtin_array
@pytest.mark.parametrize(
    ("col", "value"),
    [
        param(
            "x",
            1,
            marks=[
                pytest.mark.notimpl(
                    ["flink"],
                    raises=Py4JJavaError,
                    reason="unknown; NPE during execution",
                )
            ],
        ),
        ("y", "a"),
    ],
)
def test_array_contains(backend, con, col, value):
    t = backend.array_types
    expr = t[col].contains(value)
    result = con.execute(expr)
    expected = t[col].execute().map(lambda lst: value in lst)
    assert frozenset(result.values) == frozenset(expected.values)


@builtin_array
@pytest.mark.parametrize(
    ("a", "expected_array"),
    [
        param(
            [[1], [], [42, 42], []],
            [-1, -1, 0, -1],
            id="some-empty",
            marks=[
                pytest.mark.notyet(
                    ["flink"],
                    raises=Py4JJavaError,
                    reason="SQL validation failed; Flink does not support ARRAY[]",  # https://issues.apache.org/jira/browse/FLINK-20578
                ),
                pytest.mark.notyet(
                    ["datafusion"],
                    raises=Exception,
                    reason="Internal error: start_from index out of bounds",
                ),
            ],
        ),
        param(
            [[1], [1], [42, 42], [1]],
            [-1, -1, 0, -1],
            id="none-empty",
        ),
    ],
)
@pytest.mark.notimpl(["polars"], raises=com.OperationNotDefinedError)
@pytest.mark.notyet(["impala"], raises=com.UnsupportedBackendType)
def test_array_position(con, a, expected_array):
    t = ibis.memtable({"a": a})
    expr = t.a.index(42)
    result = con.execute(expr)
    expected = pd.Series(expected_array, dtype="object")
    assert frozenset(result.values) == frozenset(expected.values)


@builtin_array
@pytest.mark.parametrize(
    ("input", "expected"),
    [
        param(
            [[3, 2], [], [42, 2], [2, 2], []],
            [[3], [], [42], [], []],
            id="including-empty-array",
            marks=[
                pytest.mark.notyet(
                    ["flink"],
                    raises=Py4JJavaError,
                    # https://issues.apache.org/jira/browse/FLINK-20578
                    reason="SQL validation failed; Flink does not support empty array construction",
                )
            ],
        ),
        param(
            [[3, 2], [2], [42, 2], [2, 2], [2]],
            [[3], [], [42], [], []],
            id="all-non-empty-arrays",
        ),
        param(
            [[3, 2, None], [None], [42, 2], [2, 2]],
            [[3, None], [None], [42], []],
            id="nested-null",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="bigquery does not support arrays with nulls",
                ),
            ],
        ),
        param(
            [[3, 2, None], [None], [42, 2], [2, 2], None],
            [[3, None], [None], [42], [], None],
            id="non-nested-null",
            marks=[
                pytest.mark.notyet(
                    ["clickhouse"],
                    raises=AssertionError,
                    reason="clickhouse does not support nullable nested types",
                ),
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="bigquery does not support arrays with nulls",
                ),
            ],
        ),
    ],
)
def test_array_remove(con, input, expected):
    t = ibis.memtable({"a": input})
    expr = t.a.remove(2)
    result = con.execute(expr)

    lhs = frozenset(
        # arg, things are coming back as nan
        tuple(None if el is not None and math.isnan(el) else el for el in v)
        if v is not None
        else None
        for v in result.values
    )
    rhs = frozenset(tuple(v) if v is not None else None for v in expected)
    assert lhs == rhs


@builtin_array
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array..."
)
@pytest.mark.notyet(
    ["bigquery"],
    raises=GoogleBadRequest,
    reason="BigQuery doesn't support arrays with null elements",
)
@pytest.mark.notyet(
    ["clickhouse"],
    raises=(AssertionError, TypeError),
    reason="clickhouse doesn't support nullable array types",
)
@pytest.mark.notyet(
    ["bigquery"],
    raises=(AssertionError, GoogleBadRequest),
    reason="bigquery doesn't support null elements in arrays",
)
@pytest.mark.notyet(
    ["flink"], raises=Py4JJavaError, reason="empty arrays not supported"
)
@pytest.mark.notyet(
    ["datafusion"],
    raises=Exception,
    reason="arrays with NaN returns a different number of rows than expected",
)
@pytest.mark.parametrize(
    ("input", "expected"),
    [
        param(
            {"a": [[1, 3, 3], [1, 3, None, 3], [42, 42], [], [None], None]},
            [{3, 1}, {1, 3, None}, {42}, set(), {None}, None],
            id="null",
            marks=[
                pytest.mark.notyet(
                    ["polars"],
                    raises=AssertionError,
                    reason="Null elements are transformed to NaN",
                ),
                pytest.mark.notyet(
                    ["pyspark"],
                    condition=IS_SPARK_REMOTE,
                    raises=AssertionError,
                    reason="somehow, transformed results are different types",
                ),
                pytest.mark.notimpl(
                    ["databricks"], raises=AssertionError, reason="nulls are nans"
                ),
            ],
        ),
        param(
            {"a": [[1, 3, 3], [], [42, 42], [], None]},
            [{3, 1}, set(), {42}, set(), None],
            id="not_null",
        ),
    ],
)
def test_array_unique(con, input, expected):
    t = ibis.memtable(input)
    expr = t.a.unique()
    result = con.execute(expr).map(frozenset, na_action="ignore")
    expected = pd.Series(expected, dtype="object").map(frozenset, na_action="ignore")
    assert frozenset(result.values) == frozenset(expected.values)


@builtin_array
@pytest.mark.parametrize(
    "data",
    (
        param(
            [[3, 2], [], [42, 42], []],
            marks=[
                pytest.mark.notyet(
                    ["flink"],
                    raises=Py4JJavaError,
                    reason="flink cannot handle empty arrays",
                ),
            ],
            id="empty",
        ),
        param([[3, 2], [42, 42]], id="nonempty"),
    ),
)
def test_array_sort(con, data):
    t = ibis.memtable({"a": data, "id": range(len(data))})
    expr = t.mutate(a=t.a.sort()).order_by("id")
    result = con.execute(expr)
    expected = pd.Series(list(map(sorted, data)), dtype="object")

    assert frozenset(map(tuple, result["a"].values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.parametrize(
    ("a", "b", "expected_array"),
    [
        param(
            [[3, 2], [], []],
            [[1, 3], [None], [5]],
            [{1, 2, 3}, {None}, {5}],
            id="including-empty-array",
            marks=[
                pytest.mark.notyet(
                    ["flink"],
                    raises=Py4JJavaError,
                    reason="SQL validation failed; Flink does not support ARRAY[]",
                ),
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="BigQuery doesn't support arrays with null elements",
                ),
                pytest.mark.notyet(
                    ["datafusion", "polars"],
                    raises=AssertionError,
                    reason="Null elements are transformed to NaN",
                ),
                pytest.mark.notyet(
                    ["pyspark"],
                    condition=IS_SPARK_REMOTE,
                    raises=AssertionError,
                    reason="somehow, transformed results are different types",
                ),
                pytest.mark.notimpl(
                    ["databricks"], raises=AssertionError, reason="nulls are nans"
                ),
            ],
        ),
        param(
            [[3, 2], [1], [5]],
            [[1, 3], [1], [5]],
            [{1, 2, 3}, {1}, {5}],
            id="all-non-empty-arrays",
        ),
    ],
)
def test_array_union(con, a, b, expected_array):
    t = ibis.memtable({"a": a, "b": b})
    expr = t.a.union(t.b)
    result = con.execute(expr).map(set, na_action="ignore")
    expected = pd.Series(expected_array, dtype="object")

    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(["flink"], raises=com.OperationNotDefinedError)
@pytest.mark.notimpl(
    ["sqlite"], raises=com.UnsupportedBackendType, reason="Unsupported type: Array..."
)
@pytest.mark.notimpl(
    ["risingwave"],
    raises=AssertionError,
    reason="TODO(Kexiang): seems a bug",
)
@pytest.mark.parametrize(
    "data",
    [
        param(
            {"a": [[3, 2], [], []], "b": [[1, 3], [None], [5]], "c": range(3)},
            id="nulls",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="BigQuery doesn't support arrays with null elements",
                )
            ],
        ),
        param(
            {"a": [[3, 2], [], []], "b": [[1, 3], [], [5]], "c": range(3)},
            id="no_nulls",
        ),
    ],
)
def test_array_intersect(con, data):
    t = ibis.memtable(data)
    expr = t.select("c", d=t.a.intersect(t.b)).order_by("c").drop("c").d
    result = con.execute(expr).map(set, na_action="ignore")
    expected = pd.Series([{3}, set(), set()], dtype="object")
    assert len(result) == len(expected)

    assert frozenset(map(tuple, result.values)) == frozenset(
        map(tuple, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(["postgres"], raises=PsycoPg2SyntaxError)
@pytest.mark.notimpl(["risingwave"], raises=PsycoPg2InternalError)
@pytest.mark.notimpl(
    ["trino"], reason="inserting maps into structs doesn't work", raises=TrinoUserError
)
def test_unnest_struct(con):
    data = {"value": [[{"a": 1}, {"a": 2}], [{"a": 3}, {"a": 4}]]}
    t = ibis.memtable(data, schema=ibis.schema({"value": "!array<!struct<a: !int>>"}))
    expr = t.value.unnest()

    result = con.execute(expr)

    expected = pd.DataFrame(data).explode("value").iloc[:, 0].reset_index(drop=True)
    assert frozenset(map(frozendict, result.values)) == frozenset(
        map(frozendict, expected.values)
    )


@builtin_array
@pytest.mark.notimpl(["postgres"], raises=PsycoPg2SyntaxError)
@pytest.mark.notimpl(["risingwave"], raises=PsycoPg2InternalError)
@pytest.mark.notimpl(
    ["trino"], reason="inserting maps into structs doesn't work", raises=TrinoUserError
)
@pytest.mark.notimpl(
    ["flink"], reason="flink unnests a and b as separate columns", raises=Py4JJavaError
)
def test_unnest_struct_with_multiple_fields(con):
    data = {
        "value": [
            [{"a": 1, "b": "banana"}, {"a": 2, "b": "apple"}],
            [{"a": 3, "b": "coconut"}, {"a": 4, "b": "orange"}],
        ]
    }
    t = ibis.memtable(
        data, schema=ibis.schema({"value": "!array<!struct<a: !int, b: !string>>"})
    )
    expr = t.value.unnest()

    result = con.execute(expr)

    expected = pd.DataFrame(data).explode("value").iloc[:, 0].reset_index(drop=True)
    tm.assert_series_equal(result, expected)


array_zip_notimpl = pytest.mark.notimpl(
    ["datafusion", "druid", "oracle", "polars", "postgres", "risingwave", "flink"],
    raises=com.OperationNotDefinedError,
)


@builtin_array
@array_zip_notimpl
def test_zip(backend):
    t = backend.array_types

    x = t.x.execute()
    res = t.x.zip(t.x)
    assert res.type().value_type.names == ("f1", "f2")
    s = res.execute()
    assert len(s[0][0]) == len(res.type().value_type)
    assert len(x[0]) == len(s[0])

    x = t.x.execute()
    res = t.x.zip(t.x, t.x, t.x, t.x, t.x)
    assert res.type().value_type.names == ("f1", "f2", "f3", "f4", "f5", "f6")
    s = res.execute()
    assert len(s[0][0]) == len(res.type().value_type)
    assert len(x[0]) == len(s[0])


@builtin_array
@array_zip_notimpl
@pytest.mark.notyet(
    "clickhouse",
    raises=ClickHouseDatabaseError,
    reason="clickhouse nested types can't be null",
)
@pytest.mark.never(
    "bigquery",
    raises=AssertionError,
    reason="BigQuery converts NULLs with array type to an empty array",
)
@pytest.mark.parametrize(
    "fn",
    [
        param(lambda a, b: a.zip(b), id="non-null-zip-null"),
        param(lambda a, b: b.zip(a), id="null-zip-non-null"),
        param(lambda _, b: b.zip(b), id="null-zip-null"),
    ],
)
def test_zip_null(con, fn):
    a = ibis.literal([1, 2, 3], type="array<int64>")
    b = ibis.literal(None, type="array<int64>")
    expr = fn(a, b)
    assert con.execute(expr) is None


@builtin_array
@pytest.mark.notimpl(["postgres"], raises=PsycoPg2SyntaxError)
@pytest.mark.notimpl(["risingwave"], raises=PsycoPg2ProgrammingError)
@pytest.mark.notimpl(["datafusion"], raises=Exception, reason="not yet supported")
@pytest.mark.notimpl(
    ["polars"],
    raises=com.OperationNotDefinedError,
    reason="polars unnest cannot be compiled outside of a projection",
)
@pytest.mark.notyet(
    ["pyspark"],
    reason="pyspark doesn't seem to support field selection on explode",
    raises=PySparkAnalysisException,
)
@pytest.mark.notimpl(
    ["databricks"],
    reason="databricks supports about 4 ways to explode, and "
    "sqlglot doesn't implement the one that would enable this operation",
    raises=DatabricksServerOperationError,
)
@pytest.mark.notimpl(
    ["trino"], reason="inserting maps into structs doesn't work", raises=TrinoUserError
)
@pytest.mark.notyet(
    ["flink"],
    raises=Py4JJavaError,
    reason="does not seem to support field selection on unnest",
)
def test_array_of_struct_unnest(con):
    jobs = ibis.memtable(
        {
            "steps": [
                [
                    {"status": "success"},
                    {"status": "success"},
                    {"status": None},
                    {"status": "failure"},
                ],
                [
                    {"status": None},
                    {"status": "success"},
                ],
            ]
        },
        schema=dict(steps="array<struct<status: string>>"),
    )
    expr = jobs.limit(1).steps.unnest().status
    res = con.execute(expr)
    value = res.iat[0]
    # `value` can be `None` because the order of results is arbitrary; observed
    # in the wild with the trino backend
    assert value is None or isinstance(value, str)


@pytest.fixture(scope="module")
def flatten_data():
    return {
        "empty": {"data": [[], [], []], "type": "array<!array<!int64>>"},
        "happy": {
            "data": [[["abc"]], [["bcd"]], [["def"]]],
            "type": "array<!array<!string>>",
        },
        "nulls_only": {"data": [None, None, None], "type": "array<array<string>>"},
        "mixed_nulls": {"data": [[[]], None, [[None]]], "type": "array<array<string>>"},
    }


@pytest.mark.notyet(
    ["bigquery"], reason="BigQuery doesn't support arrays of arrays", raises=TypeError
)
@pytest.mark.notyet(
    ["postgres", "risingwave"],
    reason="Postgres doesn't truly support arrays of arrays",
    raises=(
        com.OperationNotDefinedError,
        PsycoPg2IndeterminateDatatype,
        PsycoPg2InternalError,
    ),
)
@pytest.mark.parametrize(
    ("column", "expected"),
    [
        param("empty", pd.Series([[], [], []], dtype="object"), id="empty"),
        param(
            "happy", pd.Series([["abc"], ["bcd"], ["def"]], dtype="object"), id="happy"
        ),
        param(
            "nulls_only",
            pd.Series([None, None, None], dtype="object"),
            id="nulls_only",
            marks=[
                pytest.mark.notyet(
                    ["clickhouse"],
                    reason="Arrays are never nullable",
                    raises=AssertionError,
                ),
            ],
        ),
        param(
            "mixed_nulls",
            pd.Series([[], None, [None]], dtype="object"),
            id="mixed_nulls",
            marks=[
                pytest.mark.notyet(
                    ["clickhouse"],
                    reason="Arrays are never nullable",
                    raises=AssertionError,
                ),
                pytest.mark.notyet(
                    ["polars"],
                    reason="flattened empty arrays incorrectly insert a null",
                    raises=AssertionError,
                ),
            ],
        ),
    ],
)
@pytest.mark.notimpl(["flink"], raises=com.OperationNotDefinedError)
def test_array_flatten(backend, flatten_data, column, expected):
    data = flatten_data[column]
    ids = range(len(data["data"]))
    t = ibis.memtable(
        {column: data["data"], "id": ids}, schema={column: data["type"], "id": "int64"}
    )
    expr = t.select("id", flat=t[column].flatten()).order_by("id")
    result = backend.connection.to_pandas(expr)
    tm.assert_frame_equal(
        result, expected.rename("flat").to_frame().assign(id=ids)[["id", "flat"]]
    )


@pytest.mark.notyet(
    ["datafusion"],
    reason="range isn't implemented upstream",
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notimpl(["flink"], raises=com.OperationNotDefinedError)
@pytest.mark.parametrize("n", [-2, 0, 2])
def test_range_single_argument(con, n):
    expr = ibis.range(n)
    result = con.execute(expr)
    assert list(result) == list(range(n))


@pytest.mark.notyet(
    ["datafusion"],
    reason="range and unnest aren't implemented upstream",
    raises=com.OperationNotDefinedError,
)
@pytest.mark.parametrize("n", [-2, 0, 2])
@pytest.mark.notimpl(["polars", "flink"], raises=com.OperationNotDefinedError)
def test_range_single_argument_unnest(con, n):
    expr = ibis.range(n).unnest()
    result = con.execute(expr)
    assert frozenset(result.values) == frozenset(range(n))


@pytest.mark.parametrize("step", [-2, -1, 1, 2])
@pytest.mark.parametrize(
    ("start", "stop"),
    [
        param(-7, -7),
        param(-7, 0),
        param(-7, 7),
        param(0, -7),
        param(0, 0),
        param(0, 7),
        param(7, -7),
        param(7, 0),
        param(7, 7),
    ],
)
@pytest.mark.notyet(
    ["datafusion"],
    reason="range and unnest aren't implemented upstream",
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notimpl(["flink"], raises=com.OperationNotDefinedError)
def test_range_start_stop_step(con, start, stop, step):
    expr = ibis.range(start, stop, step)
    result = con.execute(expr)
    assert list(result) == list(range(start, stop, step))


@pytest.mark.parametrize("stop", [-7, 0, 7])
@pytest.mark.parametrize("start", [-7, 0, 7])
@pytest.mark.notyet(
    ["clickhouse"], raises=ClickHouseDatabaseError, reason="not supported upstream"
)
@pytest.mark.notyet(
    ["datafusion"], raises=com.OperationNotDefinedError, reason="not supported upstream"
)
@pytest.mark.notimpl(["flink"], raises=com.OperationNotDefinedError)
@pytest.mark.never(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="Invalid parameter step: step size cannot equal zero",
)
def test_range_start_stop_step_zero(con, start, stop):
    expr = ibis.range(start, stop, 0)
    result = con.execute(expr)
    assert list(result) == []


@pytest.mark.notimpl(
    ["polars"],
    raises=AssertionError,
    reason="ibis hasn't implemented this behavior yet",
)
@pytest.mark.notyet(
    ["flink"],
    raises=Py4JJavaError,
    reason="SQL validation failed; Flink does not support ARRAY[]",  # https://issues.apache.org/jira/browse/FLINK-20578
)
def test_unnest_empty_array(con):
    t = ibis.memtable({"arr": [[], ["a"], ["a", "b"]]})
    expr = t.arr.unnest()
    result = con.execute(expr)
    assert len(result) == 3


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(["sqlite"], raises=com.UnsupportedBackendType)
@pytest.mark.notyet(
    "risingwave",
    raises=PsycoPg2InternalError,
    reason="no support for not null column constraint",
)
def test_array_map_with_conflicting_names(backend, con):
    t = ibis.memtable({"x": [[1, 2]]}, schema=ibis.schema(dict(x="!array<int8>")))
    expr = t.select(a=t.x.map(lambda x: x + 1)).select(
        b=lambda t: t.a.filter(lambda a: a > 2)
    )
    result = con.execute(expr)
    expected = pd.DataFrame({"b": [[3]]})
    backend.assert_frame_equal(result, expected)


@builtin_array
@pytest.mark.notimpl(
    ["datafusion", "flink", "polars", "sqlite", "sqlite"],
    raises=com.OperationNotDefinedError,
)
def test_complex_array_map(con):
    def upper(token):
        return token.upper()

    def swap(token):
        return token.substitute({"abc": "ABC"})

    arr = ibis.array(["abc", "xyz"])

    expr = arr.map(upper)
    assert con.execute(expr) == ["ABC", "XYZ"]

    expr = arr.map(swap)
    assert con.execute(expr) == ["ABC", "xyz"]

    expr = arr.map(lambda token: token.substitute({"abc": "ABC"}))
    assert con.execute(expr) == ["ABC", "xyz"]


timestamp_range_tzinfos = pytest.mark.parametrize(
    "tzinfo",
    [
        param(
            None,
            id="none",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=com.IbisTypeError,
                    reason="bigquery doesn't support datetime ranges, only timestamp ranges",
                ),
            ],
        ),
        param(
            pytz.UTC,
            id="utc",
            marks=[
                pytest.mark.notyet(
                    ["trino"],
                    raises=TrinoUserError,
                    reason="trino doesn't support timestamp with time zone arguments to its sequence function",
                ),
                pytest.mark.notyet(
                    ["polars"],
                    raises=(TypeError, com.UnsupportedOperationError),
                    reason="polars doesn't work with dateutil timezones",
                ),
            ],
        ),
    ],
)


@pytest.mark.parametrize(
    ("start", "stop", "step", "freq"),
    [
        param(
            datetime(2017, 1, 1),
            datetime(2017, 1, 2),
            ibis.interval(hours=1),
            "1h",
            id="pos",
            marks=pytest.mark.notimpl(
                ["risingwave"],
                raises=PsycoPg2InternalError,
                reason="function make_interval() does not exist",
            ),
        ),
        param(
            datetime(2017, 1, 2),
            datetime(2017, 1, 1),
            ibis.interval(hours=-1),
            "-1h",
            id="neg_inner",
            marks=[
                pytest.mark.notimpl(
                    ["polars"],
                    raises=(AssertionError, TypeError),
                    reason="returns an empty array",
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function neg(interval) does not exist",
                ),
            ],
        ),
        param(
            datetime(2017, 1, 2),
            datetime(2017, 1, 1),
            -ibis.interval(hours=1),
            "-1h",
            id="neg_outer",
            marks=[
                pytest.mark.notyet(["polars"], raises=com.UnsupportedOperationError),
                pytest.mark.notyet(["bigquery"], raises=GoogleBadRequest),
                pytest.mark.notyet(
                    ["clickhouse", "snowflake"],
                    raises=com.UnsupportedOperationError,
                ),
                pytest.mark.notimpl(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function neg(interval) does not exist",
                ),
            ],
        ),
    ],
)
@timestamp_range_tzinfos
@pytest.mark.notimpl(["flink", "datafusion"], raises=com.OperationNotDefinedError)
def test_timestamp_range(con, start, stop, step, freq, tzinfo):
    start = start.replace(tzinfo=tzinfo)
    stop = stop.replace(tzinfo=tzinfo)
    expr = ibis.range(start, stop, step)
    result = con.execute(expr)
    expected = pd.date_range(start, stop, freq=freq, inclusive="left")
    assert list(result) == expected.tolist()


@pytest.mark.parametrize(
    ("start", "stop", "step"),
    [
        param(
            datetime(2017, 1, 1, tzinfo=pytz.UTC),
            datetime(2017, 1, 2, tzinfo=pytz.UTC),
            ibis.interval(hours=0),
            id="pos",
            marks=[
                pytest.mark.notyet(["polars"], raises=PolarsComputeError),
                pytest.mark.notyet(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function make_interval() does not exist",
                ),
            ],
        ),
        param(
            datetime(2017, 1, 1, tzinfo=pytz.UTC),
            datetime(2017, 1, 2, tzinfo=pytz.UTC),
            -ibis.interval(hours=0),
            id="neg",
            marks=[
                pytest.mark.notyet(["polars"], raises=com.UnsupportedOperationError),
                pytest.mark.notyet(["bigquery"], raises=GoogleBadRequest),
                pytest.mark.notyet(
                    ["clickhouse", "snowflake"],
                    raises=com.UnsupportedOperationError,
                ),
                pytest.mark.notyet(
                    ["risingwave"],
                    raises=PsycoPg2InternalError,
                    reason="function neg(interval) does not exist",
                ),
            ],
        ),
    ],
)
@timestamp_range_tzinfos
@pytest.mark.notimpl(["flink", "datafusion"], raises=com.OperationNotDefinedError)
def test_timestamp_range_zero_step(con, start, stop, step, tzinfo):
    start = start.replace(tzinfo=tzinfo)
    stop = stop.replace(tzinfo=tzinfo)
    expr = ibis.range(start, stop, step)
    result = con.execute(expr)
    assert list(result) == []


@pytest.mark.notimpl(
    ["impala"], raises=AssertionError, reason="backend doesn't support arrays"
)
def test_repr_timestamp_array(con, monkeypatch):
    monkeypatch.setattr(ibis.options, "interactive", True)
    assert ibis.options.interactive is True

    monkeypatch.setattr(ibis.options, "default_backend", con)
    assert ibis.options.default_backend is con

    expr = ibis.array(pd.date_range("2010-01-01", "2010-01-03", freq="D").tolist())
    assert "Translation to backend failed" not in repr(expr)


@pytest.mark.notyet(
    ["datafusion", "flink", "polars"],
    raises=com.OperationNotDefinedError,
)
def test_unnest_range(con):
    expr = ibis.range(2).unnest().name("x").as_table().mutate({"y": 1.0})
    result = con.execute(expr)
    expected = pd.DataFrame({"x": np.array([0, 1], dtype="int8"), "y": [1.0, 1.0]})
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        param([1, ibis.literal(2)], [1, 2], id="int-int"),
        param([1.0, ibis.literal(2)], [1.0, 2.0], id="float-int"),
        param([1.0, ibis.literal(2.0)], [1.0, 2.0], id="float-float"),
        param([1, ibis.literal(2.0)], [1.0, 2.0], id="int-float"),
        param([ibis.literal(1), ibis.literal(2.0)], [1.0, 2.0], id="int-float-exprs"),
        param(
            [[1], ibis.literal([2])],
            [[1], [2]],
            id="array",
            marks=[pytest.mark.notyet(["bigquery"], raises=GoogleBadRequest)],
        ),
    ],
)
def test_array_literal_with_exprs(con, input, expected):
    expr = ibis.array(input)
    assert expr.op().shape == ds.scalar
    result = list(con.execute(expr))
    assert result == expected


@pytest.mark.notimpl(
    ["datafusion", "postgres", "polars", "risingwave", "flink"],
    raises=com.OperationNotDefinedError,
)
@pytest.mark.notimpl(
    ["trino"],
    raises=TrinoUserError,
    reason="sqlglot generates code that assumes there's only at most two fields to unpack from a struct",
)
def test_zip_unnest_lift(con):
    data = pd.DataFrame(dict(array1=[[1, 2, 3]], array2=[[4, 5, 6]]))
    t = ibis.memtable(data)
    zipped = t.mutate(zipped=t.array1.zip(t.array2))
    unnested = zipped.mutate(unnest=zipped.zipped.unnest())
    lifted = unnested["unnest"].lift()
    result = con.execute(lifted)
    expected = pd.DataFrame({"f1": [1, 2, 3], "f2": [4, 5, 6]})
    tm.assert_frame_equal(result, expected)


@pytest.mark.notimpl(
    ["datafusion", "polars", "flink"], raises=com.OperationNotDefinedError
)
@pytest.mark.parametrize(
    "colspec",
    ["y", lambda t: t.y, ibis._.y],
    ids=["string", "lambda", "deferred"],
)
def test_table_unnest(backend, colspec):
    t = backend.array_types
    expr = t.unnest(colspec)
    result = expr.execute()
    assert set(result["y"].values) == set(t[["y"]].execute().explode("y")["y"].values)


@pytest.mark.notimpl(
    ["datafusion", "polars", "flink"], raises=com.OperationNotDefinedError
)
def test_table_unnest_with_offset(backend):
    t = backend.array_types
    col = "y"
    df = (
        t[[col]]
        .execute()
        .assign(idx=lambda df: df[col].map(lambda v: list(range(len(v)))))[[col, "idx"]]
        .explode("idx")
        .assign(idx=lambda df: df["idx"].astype("int64"))
    )
    idx = iter(df.idx.values)
    expected = (
        df.assign(**{col: df[col].map(lambda v: v[next(idx)])})
        .sort_values(["idx", col])
        .reset_index(drop=True)[["idx", col]]
    )

    expr = t.unnest(col, offset="idx")[["idx", col]].order_by("idx", col)
    result = expr.execute()
    tm.assert_frame_equal(result, expected)


@pytest.mark.notimpl(
    ["datafusion", "polars", "flink"], raises=com.OperationNotDefinedError
)
def test_table_unnest_with_keep_empty(con):
    t = ibis.memtable(pd.DataFrame({"y": [[], None, ["a"]]}))
    expr = t.unnest("y", keep_empty=True)["y"]
    result = con.execute(expr)
    assert Counter(result.values) == Counter(["a", None, None])


@pytest.mark.notimpl(
    ["datafusion", "polars", "flink"], raises=com.OperationNotDefinedError
)
@pytest.mark.notyet(
    ["risingwave"], raises=PsycoPg2InternalError, reason="not supported in risingwave"
)
def test_table_unnest_column_expr(backend):
    t = backend.array_types
    expr = t.unnest(t.y.map(lambda v: v.cast("str") + "'s").name("plural"))
    result = expr.execute()["plural"]
    expected = t["y"].execute().explode("y") + "'s"
    assert set(result.values) == set(expected.replace({np.nan: None}).values)


@pytest.mark.notimpl(
    ["datafusion", "polars", "flink"], raises=com.OperationNotDefinedError
)
@pytest.mark.notimpl(["trino"], raises=TrinoUserError)
@pytest.mark.notimpl(["postgres"], raises=PsycoPg2SyntaxError)
@pytest.mark.notimpl(["risingwave"], raises=PsycoPg2ProgrammingError)
@pytest.mark.notyet(
    ["risingwave"], raises=PsycoPg2InternalError, reason="not supported in risingwave"
)
def test_table_unnest_array_of_struct_of_array(con):
    t = ibis.memtable(
        {
            "a": [
                [{"x": [1, 2, 3]}, {"x": [1, 2]}],
                [],
                None,
                [{"x": [3, 1, 2, 3]}],
            ]
        },
        schema={"a": "array<struct<x: array<int64>>>"},
    )
    # two different unnests
    expr = t.unnest("a").a.x.unnest().name("x").as_table().order_by("x")
    result = con.execute(expr)
    expected = pd.DataFrame({"x": [1, 1, 1, 2, 2, 2, 3, 3, 3]})
    tm.assert_frame_equal(result, expected)


notimpl_aggs = pytest.mark.notimpl(
    ["datafusion", "flink"], raises=com.OperationNotDefinedError
)


def _agg_with_nulls(agg, x):
    if x is None:
        return None
    x = [y for y in x if not pd.isna(y)]
    if not x:
        return None
    return agg(x)


@pytest.mark.parametrize(
    ("agg", "baseline_func"),
    [
        (ir.ArrayValue.sums, lambda x: _agg_with_nulls(sum, x)),
        (ir.ArrayValue.mins, lambda x: _agg_with_nulls(min, x)),
        (ir.ArrayValue.maxs, lambda x: _agg_with_nulls(max, x)),
        (ir.ArrayValue.means, lambda x: _agg_with_nulls(statistics.mean, x)),
    ],
    ids=["sums", "mins", "maxs", "means"],
)
@notimpl_aggs
@pytest.mark.parametrize(
    "data",
    [
        param(
            [[None, 6], [None]],
            id="nulls",
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="bigquery doesn't allow arrays with nulls",
                )
            ],
        ),
        param([[1, 2, 3], [6], [], None], id="no-nulls"),
    ],
)
def test_array_agg_numeric(con, data, agg, baseline_func):
    t = ibis.memtable({"x": data, "id": range(len(data))})
    t = t.mutate(y=agg(t.x))
    assert t.y.type().is_numeric()
    # sort so debugging is easier
    df = con.to_pandas(t.order_by("id"))
    result = df.y.tolist()
    result = [x if pd.notna(x) else None for x in result]
    expected = [baseline_func(x) for x in df.x]
    assert result == expected


@pytest.mark.parametrize(
    ("agg", "baseline_func"),
    [
        (ir.ArrayValue.anys, partial(_agg_with_nulls, any)),
        (ir.ArrayValue.alls, partial(_agg_with_nulls, all)),
    ],
    ids=["anys", "alls"],
)
@pytest.mark.parametrize(
    "data",
    [
        param(
            [[True, None], [False, None], [None]],
            marks=[
                pytest.mark.notyet(
                    ["bigquery"],
                    raises=GoogleBadRequest,
                    reason="bigquery doesn't allow arrays with nulls",
                )
            ],
            id="nulls",
        ),
        param([[True, False], [True], [False], [], None], id="no-nulls"),
    ],
)
@notimpl_aggs
def test_array_agg_bool(con, data, agg, baseline_func):
    t = ibis.memtable({"x": data, "id": range(len(data))})
    t = t.mutate(y=agg(t.x))
    assert t.y.type().is_boolean()
    # sort so debugging is easier
    df = con.to_pandas(t.order_by("id"))
    result = df.y.tolist()
    result = [x if pd.notna(x) else None for x in result]
    expected = [baseline_func(x) for x in df.x]
    assert result == expected


@pytest.mark.notyet(
    ["postgres"],
    raises=PsycoPg2ArraySubscriptError,
    reason="all dimensions must match in size",
)
@pytest.mark.notimpl(["risingwave", "flink"], raises=com.OperationNotDefinedError)
@pytest.mark.notyet(
    ["bigquery"], raises=TypeError, reason="nested arrays aren't supported"
)
def test_flatten(con):
    t = ibis.memtable(
        [{"arr": [[1, 5, 7], [3, 4]]}], schema={"arr": "array<array<int64>>"}
    )
    expr = t.arr.flatten().name("result")
    result = con.execute(expr)
    expected = pd.Series([[1, 5, 7, 3, 4]], name="result")
    tm.assert_series_equal(result, expected)



from __future__ import annotations

from datetime import datetime, timedelta

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.rules as rlz
import ibis.expr.types as ir
from ibis.expr import api


def test_field_select(alltypes):
    assert isinstance(alltypes.i, ir.TimestampColumn)


def test_string_cast_to_timestamp(alltypes):
    casted = alltypes.g.cast("timestamp")
    assert isinstance(casted, ir.TimestampColumn)

    string = api.literal("2000-01-01")
    casted = string.cast("timestamp")
    assert isinstance(casted, ir.TimestampScalar)


@pytest.mark.parametrize(
    ("field", "expected_operation", "expected_type"),
    [
        ("year", ops.ExtractYear, ir.IntegerColumn),
        ("month", ops.ExtractMonth, ir.IntegerColumn),
        ("day", ops.ExtractDay, ir.IntegerColumn),
        ("hour", ops.ExtractHour, ir.IntegerColumn),
        ("minute", ops.ExtractMinute, ir.IntegerColumn),
        ("second", ops.ExtractSecond, ir.IntegerColumn),
        ("microsecond", ops.ExtractMicrosecond, ir.IntegerColumn),
        ("millisecond", ops.ExtractMillisecond, ir.IntegerColumn),
    ],
)
def test_extract_fields(field, expected_operation, expected_type, alltypes):
    # type-size may be database specific
    result = getattr(alltypes.i, field)()
    assert isinstance(result, expected_type)
    assert isinstance(result.op(), expected_operation)


def test_now():
    result = api.now()
    assert isinstance(result, ir.TimestampScalar)
    assert isinstance(result.op(), ops.TimestampNow)


def test_timestamp_literals():
    expr = ibis.timestamp("2015-01-01 00:00:00")
    assert isinstance(expr, ir.TimestampScalar)


def test_invalid_timestamp_literal():
    with pytest.raises(TypeError, match="Unable to normalize"):
        ibis.timestamp("2015-01-01 00:71")


def test_comparison_timestamp(alltypes):
    expr = alltypes.i > alltypes.i.min() + ibis.interval(days=3)
    assert isinstance(expr, ir.BooleanColumn)


def test_comparisons_string(alltypes):
    val = "2015-01-01 00:00:00"
    expr = alltypes.i > val
    op = expr.op()
    assert op.right.dtype is dt.string

    expr2 = val < alltypes.i
    op = expr2.op()
    assert isinstance(op, ops.Greater)
    assert op.right.dtype is dt.string


def test_comparisons_pandas_timestamp(alltypes):
    pd = pytest.importorskip("pandas")

    val = pd.Timestamp("2015-01-01 00:00:00")
    expr = alltypes.i > val
    op = expr.op()
    assert isinstance(op, ops.Greater)
    assert isinstance(op.right, ops.Literal)
    assert isinstance(op.right.dtype, dt.Timestamp)

    expr = ibis.literal(val) < alltypes.i
    op = expr.op()
    assert isinstance(op, ops.Less)
    assert isinstance(op.left, ops.Literal)
    assert isinstance(op.left.dtype, dt.Timestamp)

    expr = val < alltypes.i
    op = expr.op()
    assert isinstance(op, ops.Greater)
    assert isinstance(op.right, ops.Literal)
    assert isinstance(op.right.dtype, dt.Timestamp)


def test_timestamp_precedence():
    ts = ibis.literal(datetime.now())
    highest_type = rlz.highest_precedence_dtype([ibis.null().op(), ts.op()])
    assert highest_type == dt.timestamp


@pytest.mark.parametrize(
    ("field", "expected_operation", "expected_type"),
    [
        ("year", ops.ExtractYear, ir.IntegerColumn),
        ("month", ops.ExtractMonth, ir.IntegerColumn),
        ("day", ops.ExtractDay, ir.IntegerColumn),
    ],
)
def test_timestamp_field_access_on_date(
    field, expected_operation, expected_type, alltypes
):
    date_col = alltypes.i.date()
    result = getattr(date_col, field)()
    assert isinstance(result, expected_type)
    assert isinstance(result.op(), expected_operation)


@pytest.mark.parametrize(
    ("field", "expected_operation", "expected_type"),
    [
        ("hour", ops.ExtractHour, ir.IntegerColumn),
        ("minute", ops.ExtractMinute, ir.IntegerColumn),
        ("second", ops.ExtractSecond, ir.IntegerColumn),
        ("microsecond", ops.ExtractMicrosecond, ir.IntegerColumn),
        ("millisecond", ops.ExtractMillisecond, ir.IntegerColumn),
    ],
)
def test_timestamp_field_access_on_date_failure(
    field, expected_operation, expected_type, alltypes
):
    time_col = alltypes.i.date()
    with pytest.raises(AttributeError):
        getattr(time_col, field)


@pytest.mark.parametrize(
    ("field", "expected_operation", "expected_type"),
    [
        ("hour", ops.ExtractHour, ir.IntegerColumn),
        ("minute", ops.ExtractMinute, ir.IntegerColumn),
        ("second", ops.ExtractSecond, ir.IntegerColumn),
        ("microsecond", ops.ExtractMicrosecond, ir.IntegerColumn),
        ("millisecond", ops.ExtractMillisecond, ir.IntegerColumn),
    ],
)
def test_timestamp_field_access_on_time(
    field, expected_operation, expected_type, alltypes
):
    time_col = alltypes.i.time()
    result = getattr(time_col, field)()
    assert isinstance(result, expected_type)
    assert isinstance(result.op(), expected_operation)


@pytest.mark.parametrize("field", ["year", "month", "day"])
def test_timestamp_field_access_on_time_failure(field, alltypes):
    date_col = alltypes.i.time()
    with pytest.raises(AttributeError):
        getattr(date_col, field)


def test_integer_timestamp_fails():
    with pytest.raises(TypeError, match=r"Use ibis\.literal\(\.\.\.\)\.as_timestamp"):
        ibis.timestamp(42)


def test_to_timestamp_deprecation():
    with pytest.warns(FutureWarning, match="v10.0"):
        ibis.literal(42).to_timestamp()


@pytest.mark.parametrize(
    "start",
    [
        "2002-01-01 00:00:00",
        datetime(2002, 1, 1, 0, 0, 0),
        ibis.timestamp("2002-01-01 00:00:00"),
        ibis.timestamp(datetime(2002, 1, 1, 0, 0, 0)),
        ibis.table({"start": "timestamp"}).start,
    ],
)
@pytest.mark.parametrize(
    "stop",
    [
        "2002-01-02 00:00:00",
        datetime(2002, 1, 2, 0, 0, 0),
        ibis.timestamp("2002-01-02 00:00:00"),
        ibis.timestamp(datetime(2002, 1, 2, 0, 0, 0)),
        ibis.table({"stop": "timestamp"}).stop,
    ],
)
@pytest.mark.parametrize("step", [ibis.interval(seconds=1), timedelta(seconds=1)])
def test_timestamp_range_with_str_inputs(start, stop, step):
    expr = ibis.range(start, stop, step)

    op = expr.op()

    assert op.start.dtype.is_timestamp()
    assert op.stop.dtype.is_timestamp()
    assert op.step.dtype.is_interval()

    dtype = expr.type()

    assert dtype.is_array()
    assert dtype.value_type.is_timestamp()



from __future__ import annotations

import contextlib
import datetime
import re
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import pyexasol
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateDatabase
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import STAR, C

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa

    from ibis.backends import BaseBackend

# strip trailing encodings e.g., UTF8
_VARCHAR_REGEX = re.compile(r"^((VAR)?CHAR(?:\(\d+\)))?(?:\s+.+)?$")


class Backend(SQLBackend, CanCreateDatabase):
    name = "exasol"
    compiler = sc.exasol.compiler
    supports_temporary_tables = False
    supports_create_or_replace = False
    supports_python_udfs = False

    @property
    def version(self) -> str:
        # https://stackoverflow.com/a/67500385
        query = (
            sg.select("param_value")
            .from_(sg.table("EXA_METADATA", catalog="SYS"))
            .where(C.param_name.eq(sge.convert("databaseProductVersion")))
        )
        with self._safe_raw_sql(query) as result:
            [(version,)] = result.fetchall()
        return version

    def do_connect(
        self,
        user: str,
        password: str,
        host: str = "localhost",
        port: int = 8563,
        timezone: str = "UTC",
        **kwargs: Any,
    ) -> None:
        """Create an Ibis client connected to an Exasol database.

        Parameters
        ----------
        user
            Username used for authentication.
        password
            Password used for authentication.
        host
            Hostname to connect to.
        port
            Port number to connect to.
        timezone
            The session timezone.
        kwargs
            Additional keyword arguments passed to `pyexasol.connect`.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_EXASOL_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_EXASOL_USER", "sys")
        >>> password = os.environ.get("IBIS_TEST_EXASOL_PASSWORD", "exasol")
        >>> schema = os.environ.get("IBIS_TEST_EXASOL_DATABASE", "EXASOL")
        >>> con = ibis.exasol.connect(schema=schema, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int32
          bool_col        boolean
          tinyint_col     int16
          smallint_col    int16
          int_col         int32
          bigint_col      int64
          float_col       float64
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp(3)
          year            int32
          month           int32
        """
        if kwargs.pop("quote_ident", None) is not None:
            raise com.UnsupportedArgumentError(
                "Setting `quote_ident` to anything other than `True` is not supported. "
                "Ibis requires all identifiers to be quoted to work correctly."
            )

        self.con = pyexasol.connect(
            dsn=f"{host}:{port}",
            user=user,
            password=password,
            quote_ident=True,
            **kwargs,
        )
        self._post_connect(timezone)

    @util.experimental
    @classmethod
    def from_connection(
        cls, con: pyexasol.ExaConnection, timezone: str | None = None
    ) -> Backend:
        """Create an Ibis client from an existing connection to an Exasol database.

        Parameters
        ----------
        con
            An existing connection to an Exasol database.
        timezone
            The session timezone.
        """
        if timezone is None:
            timezone = (con.execute("SELECT SESSIONTIMEZONE").fetchone() or ("UTC",))[0]

        new_backend = cls(timezone=timezone)
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect(timezone)
        return new_backend

    def _post_connect(self, timezone: str = "UTC") -> None:
        with self.begin() as con:
            con.execute(f"ALTER SESSION SET TIME_ZONE = {timezone!r}")

    def _from_url(self, url: ParseResult, **kwargs) -> BaseBackend:
        """Construct an ibis backend from a URL."""
        kwargs = {
            "user": url.username,
            "password": unquote_plus(url.password)
            if url.password is not None
            else None,
            "schema": url.path[1:] or None,
            "host": url.hostname,
            "port": url.port,
            **kwargs,
        }

        self._convert_kwargs(kwargs)

        return self.connect(**kwargs)

    @contextlib.contextmanager
    def begin(self):
        # pyexasol doesn't have a cursor method
        con = self.con
        try:
            yield con
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()

    @contextlib.contextmanager
    def _safe_raw_sql(self, query: str, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)

        with self.begin() as cur:
            yield cur.execute(query, *args, **kwargs)

    def list_tables(self, like=None, database=None):
        """List the tables in the database.

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Database to list tables from. Default behavior is to show tables in
            the current database.
        """
        tables = sg.select("table_name").from_(
            sg.table("EXA_ALL_TABLES", catalog="SYS")
        )
        views = sg.select(sg.column("view_name").as_("table_name")).from_(
            sg.table("EXA_ALL_VIEWS", catalog="SYS")
        )

        if database is not None:
            tables = tables.where(sg.column("table_schema").eq(sge.convert(database)))
            views = views.where(sg.column("view_schema").eq(sge.convert(database)))

        query = sg.union(tables, views)

        with self._safe_raw_sql(query) as con:
            tables = con.fetchall()

        return self._filter_with_like([table for (table,) in tables], like=like)

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        query = (
            sg.select(STAR)
            .from_(
                sg.table(
                    table_name,
                    db=database,
                    catalog=catalog,
                    quoted=self.compiler.quoted,
                )
            )
            .sql(self.dialect)
        )
        try:
            return self._get_schema_using_query(query)
        except pyexasol.exceptions.ExaQueryError as e:
            if not self.con.meta.table_exists(table_name):
                raise com.TableNotFound(table_name) from e
            raise

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        import pandas as pd

        from ibis.backends.exasol.converter import ExasolPandasData

        df = pd.DataFrame.from_records(cursor, columns=schema.names, coerce_float=True)
        df = ExasolPandasData.convert_table(df, schema)
        return df

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        table = sg.table(
            util.gen_name(f"{self.name}_metadata"), quoted=self.compiler.quoted
        )
        dialect = self.dialect
        create_view = sg.exp.Create(
            kind="VIEW",
            this=table,
            expression=sg.parse_one(query, dialect=dialect),
        )
        drop_view = sg.exp.Drop(kind="VIEW", this=table).sql(dialect)
        describe = sg.exp.Describe(this=table).sql(dialect)
        type_mapper = self.compiler.type_mapper
        with self._safe_raw_sql(create_view):
            try:
                return sch.Schema(
                    {
                        name: type_mapper.from_string(_VARCHAR_REGEX.sub(r"\1", typ))
                        for name, typ, *_ in self.con.execute(describe).fetchall()
                    }
                )
            finally:
                self.con.execute(drop_view)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "Exasol cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        quoted = self.compiler.quoted
        name = op.name

        ident = sg.to_identifier(name, quoted=quoted)
        create_stmt = sg.exp.Create(
            kind="TABLE",
            this=sg.exp.Schema(this=ident, expressions=schema.to_sqlglot(self.dialect)),
        )
        create_stmt_sql = create_stmt.sql(self.name)

        df = op.data.to_frame()
        data = df.itertuples(index=False, name=None)

        def process_item(item: Any):
            """Handle inserting timestamps with timezones."""
            if isinstance(item, datetime.datetime):
                if item.tzinfo is not None:
                    item = item.tz_convert("UTC").tz_localize(None)
                return item.isoformat(sep=" ", timespec="milliseconds")
            return item

        rows = (tuple(map(process_item, row)) for row in data)
        with self._safe_raw_sql(create_stmt_sql):
            if not df.empty:
                self.con.ext.insert_multi(name, rows)

    def _clean_up_tmp_table(self, name: str) -> None:
        ident = sg.to_identifier(name, quoted=self.compiler.quoted)
        sql = sge.Drop(kind="TABLE", this=ident, exists=True, cascade=True)
        with self._safe_raw_sql(sql):
            pass

    _finalize_memtable = _clean_up_tmp_table

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        overwrite: bool = False,
        temp: bool = False,
    ) -> ir.Table:
        """Create a table in Exasol.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The database in which to create the table; optional
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists
        temp
            Create a temporary table (not supported)

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        if temp:
            raise com.UnsupportedOperationError(
                f"Creating temp tables is not supported by {self.name}"
            )

        quoted = self.compiler.quoted

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        if not schema:
            schema = table.schema()

        table_expr = sg.table(temp_name, catalog=database, quoted=quoted)
        target = sge.Schema(
            this=table_expr, expressions=schema.to_sqlglot(self.dialect)
        )

        create_stmt = sge.Create(kind="TABLE", this=target)

        this = sg.table(name, catalog=database, quoted=quoted)
        with self._safe_raw_sql(create_stmt):
            if query is not None:
                self.con.execute(
                    sge.Insert(this=table_expr, expression=query).sql(self.name)
                )

            if overwrite:
                self.con.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.name)
                )
                self.con.execute(
                    f"RENAME TABLE {table_expr.sql(self.name)} TO {this.sql(self.name)}"
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql("SELECT CURRENT_SCHEMA") as cur:
            [(schema,)] = cur.fetchall()
        return schema

    def drop_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None:
            raise NotImplementedError(
                "`catalog` argument is not supported for the Exasol backend"
            )
        drop_schema = sg.exp.Drop(
            kind="SCHEMA",
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            exists=force,
        )
        with self.begin() as con:
            con.execute(drop_schema.sql(dialect=self.dialect))

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None:
            raise NotImplementedError(
                "`catalog` argument is not supported for the Exasol backend"
            )
        create_database = sg.exp.Create(
            kind="SCHEMA",
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            exists=force,
        )
        open_database = self.current_database
        with self.begin() as con:
            con.execute(create_database.sql(dialect=self.dialect))
            # Exasol implicitly opens the created schema, therefore we need to restore
            # the previous context.
            con.execute(
                f"OPEN SCHEMA {open_database}"
                if open_database
                else f"CLOSE SCHEMA {name}"
            )

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        if catalog is not None:
            raise NotImplementedError(
                "`catalog` argument is not supported for the Exasol backend"
            )

        query = sg.select("schema_name").from_(sg.table("EXA_SCHEMAS", catalog="SYS"))

        with self._safe_raw_sql(query) as con:
            databases = con.fetchall()
        return self._filter_with_like([db for (db,) in databases], like=like)

    def _cursor_batches(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1 << 20,
    ) -> Iterable[list]:
        self._run_pre_execute_hooks(expr)

        dtypes = expr.as_table().schema().values()

        with self._safe_raw_sql(
            self.compile(expr, limit=limit, params=params)
        ) as cursor:
            while batch := cursor.fetchmany(chunk_size):
                yield (tuple(map(dt.normalize, dtypes, row)) for row in batch)



"""Tests for json data types."""

from __future__ import annotations

import json

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir


@pytest.fixture(scope="module")
def jsonb_t(con):
    return con.table("jsonb_t")


@pytest.mark.parametrize("data", [param({"status": True}, id="status")])
def test_json(data, alltypes):
    lit = ibis.literal(json.dumps(data), type="json").name("tmp")
    expr = alltypes.select(alltypes.id, lit).head(1)
    df = expr.execute()
    assert df["tmp"].iloc[0] == data


def test_jsonb_extract_path(con):
    json_t = con.table("json_t")
    jsonb_t = con.table("jsonb_t")

    assert json_t.js.type() == dt.JSON(binary=False)
    assert jsonb_t.js.type() == dt.JSON(binary=True)

    tm.assert_series_equal(jsonb_t.js["a"].execute(), json_t.js["a"].execute())


def test_json_getitem_object(jsonb_t):
    expr_fn = lambda t: t.js["a"].name("res")
    expected = frozenset([(1, 2, 3, 4), None, "foo"] + [None] * 3)
    expr = expr_fn(jsonb_t)
    result = frozenset(
        expr.execute()
        .map(lambda o: tuple(o) if isinstance(o, list) else o)
        .replace({np.nan: None})
    )
    assert result == expected


def test_json_getitem_array(jsonb_t):
    expr_fn = lambda t: t.js[1].name("res")
    expected = frozenset([None] * 4 + [47, None])
    expr = expr_fn(jsonb_t)
    result = frozenset(expr.execute().replace({np.nan: None}))
    assert result == expected


def test_json_array(jsonb_t):
    expr = jsonb_t.mutate("rowid", res=jsonb_t.js.array).order_by("rowid")
    result = expr.execute().res
    expected = pd.Series(
        [None, None, None, None, [42, 47, 55], []] + [None] * 8,
        name="res",
        dtype="object",
    )
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("typ", "expected_data"),
    [
        ("str", [None] * 6 + ["a", "", "b"] + [None] * 5),
        ("int", [None] * 12 + [42, None]),
        ("float", [None] * 12 + [42.0, 37.37]),
        ("bool", [None] * 10 + [True, False, None, None]),
    ],
    ids=["str", "int", "float", "bool"],
)
@pytest.mark.parametrize(
    "expr_fn", [getattr, ir.JSONValue.unwrap_as], ids=["getattr", "unwrap_as"]
)
def test_json_unwrap(jsonb_t, typ, expected_data, expr_fn):
    expr = expr_fn(jsonb_t.js, typ).name("res")
    result = expr.execute()
    expected = pd.Series(expected_data, name="res", dtype="object")
    tm.assert_series_equal(
        result.replace(np.nan, None).fillna(pd.NA).sort_values().reset_index(drop=True),
        expected.replace(np.nan, None)
        .fillna(pd.NA)
        .sort_values()
        .reset_index(drop=True),
        check_dtype=False,
    )



"""Tests for JSON operations."""

from __future__ import annotations

import sqlite3

import pytest
from packaging.version import parse as vparse

import ibis.expr.types as ir
from ibis.backends.tests.errors import PySparkPythonException
from ibis.conftest import IS_SPARK_REMOTE

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")

pytestmark = [
    pytest.mark.never(["impala"], reason="doesn't support JSON and never will"),
    pytest.mark.notyet(["clickhouse"], reason="upstream is broken"),
    pytest.mark.notimpl(
        ["datafusion", "exasol", "mssql", "druid", "oracle", "databricks"]
    ),
]


@pytest.mark.notyet(
    ["sqlite"],
    condition=vparse(sqlite3.sqlite_version) < vparse("3.38.0"),
    reason="JSON not supported in SQLite < 3.38.0",
)
@pytest.mark.notimpl(
    ["flink"],
    reason="https://github.com/ibis-project/ibis/pull/6920#discussion_r1373212503",
)
@pytest.mark.notimpl(
    ["risingwave"], reason="TODO(Kexiang): order mismatch in array", strict=False
)
def test_json_getitem_object(json_t):
    expr_fn = lambda t: t.js["a"].name("res")
    expected = frozenset([(1, 2, 3, 4), None, "foo"] + [None] * 3)
    expr = expr_fn(json_t)
    result = frozenset(
        expr.execute()
        .map(lambda o: tuple(o) if isinstance(o, list) else o)
        .replace({np.nan: None})
    )
    assert result == expected


@pytest.mark.notyet(
    ["sqlite"],
    condition=vparse(sqlite3.sqlite_version) < vparse("3.38.0"),
    reason="JSON not supported in SQLite < 3.38.0",
)
@pytest.mark.notimpl(
    ["flink"],
    reason="https://github.com/ibis-project/ibis/pull/6920#discussion_r1373212503",
)
@pytest.mark.notimpl(
    ["risingwave"], reason="TODO(Kexiang): order mismatch in array", strict=False
)
def test_json_getitem_array(json_t):
    expr_fn = lambda t: t.js[1].name("res")
    expected = frozenset([None] * 4 + [47, None])
    expr = expr_fn(json_t)
    result = frozenset(expr.execute().replace({np.nan: None}))
    assert result == expected


@pytest.mark.notimpl(["mysql", "risingwave"])
@pytest.mark.notyet(["bigquery", "sqlite"], reason="doesn't support maps")
@pytest.mark.notyet(["postgres"], reason="only supports map<string, string>")
@pytest.mark.notyet(
    ["pyspark", "flink"], reason="should work but doesn't deserialize JSON"
)
def test_json_map(backend, json_t):
    expr = json_t.mutate("rowid", res=json_t.js.map).order_by("rowid")
    result = expr.execute().res
    expected = pd.Series(
        [
            {"a": [1, 2, 3, 4], "b": 1},
            {"a": None, "b": 2},
            {"a": "foo", "c": None},
        ]
        + [None] * 11,
        dtype="object",
        name="res",
    )
    backend.assert_series_equal(result, expected)


@pytest.mark.notimpl(["mysql", "risingwave"])
@pytest.mark.notyet(["sqlite"], reason="doesn't support arrays")
@pytest.mark.notyet(
    ["pyspark", "flink"], reason="should work but doesn't deserialize JSON"
)
@pytest.mark.notyet(["bigquery"], reason="doesn't allow null in arrays")
def test_json_array(backend, json_t):
    expr = json_t.mutate("rowid", res=json_t.js.array).order_by("rowid")
    result = expr.execute().res
    expected = pd.Series(
        [None, None, None, None, [42, 47, 55], []] + [None] * 8,
        name="res",
        dtype="object",
    )
    backend.assert_series_equal(result, expected)


@pytest.mark.notyet(
    ["sqlite"],
    condition=vparse(sqlite3.sqlite_version) < vparse("3.38.0"),
    reason="JSON not supported in SQLite < 3.38.0",
)
@pytest.mark.notimpl(["risingwave"])
@pytest.mark.notyet(["flink"], reason="should work but doesn't deserialize JSON")
@pytest.mark.notyet(
    ["pyspark"],
    condition=IS_SPARK_REMOTE,
    raises=PySparkPythonException,
    reason="environment issues",
)
@pytest.mark.parametrize(
    ("typ", "expected_data"),
    [
        ("str", [None] * 6 + ["a", "", "b"] + [None] * 5),
        ("int", [None] * 12 + [42, None]),
        ("float", [None] * 12 + [42.0, 37.37]),
        ("bool", [None] * 10 + [True, False, None, None]),
    ],
    ids=["str", "int", "float", "bool"],
)
@pytest.mark.parametrize(
    "expr_fn", [getattr, ir.JSONValue.unwrap_as], ids=["getattr", "unwrap_as"]
)
def test_json_unwrap(backend, json_t, typ, expected_data, expr_fn):
    expr = expr_fn(json_t.js, typ).name("res")
    result = expr.execute()
    expected = pd.Series(expected_data, name="res", dtype="object")
    backend.assert_series_equal(
        result.replace(np.nan, None).fillna(pd.NA).sort_values().reset_index(drop=True),
        expected.replace(np.nan, None)
        .fillna(pd.NA)
        .sort_values()
        .reset_index(drop=True),
        check_dtype=False,
    )



from __future__ import annotations

import datetime
import functools
import re
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import pytest
import sqlglot as sg
from dateutil.relativedelta import relativedelta

import ibis

if TYPE_CHECKING:
    from collections.abc import Callable

    import ibis.expr.types as ir


def pytest_pyfunc_call(pyfuncitem):
    """Inject `backend` and fixtures to all TPC-DS test functions.

    Defining this hook here limits its scope to the TPC-DS tests.
    """
    testfunction = pyfuncitem.obj
    funcargs = pyfuncitem.funcargs
    testargs = {arg: funcargs[arg] for arg in pyfuncitem._fixtureinfo.argnames}

    if (backend := funcargs.get("backend")) is not None:
        testargs["backend"] = backend

    result = testfunction(**testargs)
    assert (
        result is None
    ), "test function should not return anything, did you mean to use assert?"
    return True


def tpc_test(suite_name: Literal["h", "ds"], *, result_is_empty=False):
    """Decorator for TPC tests.

    Parameters
    ----------
    suite_name
        The name of the TPC suite. Only `'h'` and ~'ds'~ are supported right now.
    result_is_empty
        If the expected result is an empty table.

    Automates the process of loading the SQL query from the file system and
    asserting that the result of the ibis expression is equal to the expected
    result of executing the raw SQL.
    """

    def inner(test: Callable[..., ir.Table]):
        name = f"tpc{suite_name}"

        # so that clickhouse doesn't run forever when we hit one of its weird cross
        # join performance black holes
        #
        # trino can sometimes take a while as well, especially in CI
        #
        # func_only=True doesn't include the fixture setup time in the duration
        # of the test run, which is important since backends can take a hugely
        # variable amount of time to load all the TPC-$WHATEVER tables.
        @pytest.mark.timeout(90, func_only=True)
        @pytest.mark.usefixtures("backend")
        @pytest.mark.xdist_group(name)
        @getattr(pytest.mark, name)
        @functools.wraps(test)
        def wrapper(*args, backend, **kwargs):
            from ibis.formats.pandas import PandasData

            backend_name = backend.name()
            if not getattr(backend, f"supports_{name}"):
                pytest.skip(
                    f"{backend_name} backend doesn't support testing {name} queries yet"
                )
            query_name_match = re.match(r"^test_(\d\d)$", test.__name__)
            assert query_name_match is not None

            query_number = query_name_match.group(1)
            sql_path_name = f"{query_number}.sql"

            base = Path(__file__).parent / "queries"

            path = base / backend_name / suite_name / sql_path_name

            if path.exists():
                dialect = backend_name
            else:
                dialect = "duckdb"
                path = base / "duckdb" / suite_name / sql_path_name

            raw_sql = path.read_text()

            sql = sg.parse_one(raw_sql, read=dialect)

            sql = backend._transform_tpc_sql(
                sql, suite=suite_name, leaves=backend.list_tpc_tables(suite_name)
            )

            raw_sql = sql.sql(dialect=dialect, pretty=True)

            expected_expr = backend.connection.sql(raw_sql, dialect=dialect)

            result_expr = test(*args, **kwargs)

            assert result_expr._find_backend(use_default=False) is backend.connection
            result = backend.connection.to_pandas(result_expr)

            expected = expected_expr.to_pandas()

            assert (result_is_empty and result.empty) or (
                not result_is_empty and not result.empty
            )

            # First check that the column names match up
            assert len(expected.columns) == len(result.columns)
            assert all(
                r.lower() in e.lower() for r, e in zip(result.columns, expected.columns)
            )

            # Then set the expected columns so we can coerce the datatypes
            # of the pandas dataframe correctly
            expected.columns = result.columns

            expected = PandasData.convert_table(expected, result_expr.schema())

            # Then run the value comparisons
            compare_tpc_results(
                result,
                expected,
                abs_tol=backend.tpc_absolute_tolerance,
                result_is_empty=result_is_empty,
            )

        return wrapper

    return inner


def compare_tpc_results(result, expected, result_is_empty=False, abs_tol=0.001):
    assert (result_is_empty and expected.empty) or (
        not result_is_empty and not expected.empty
    )

    assert len(expected) == len(result)
    assert result.columns.tolist() == expected.columns.tolist()
    for column in result.columns:
        left = result.loc[:, column]
        right = expected.loc[:, column]
        assert (
            pytest.approx(
                left.values.tolist(),
                nan_ok=True,
                abs=abs_tol,
            )
            == right.values.tolist()
        )


def add_date(datestr: str, dy: int = 0, dm: int = 0, dd: int = 0) -> ir.DateScalar:
    dt = datetime.date.fromisoformat(datestr)
    dt += relativedelta(years=dy, months=dm, days=dd)
    return ibis.date(dt.isoformat())



from __future__ import annotations

from typing import Any

import pytest
import sqlglot as sg

import ibis
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.data import array_types, topk, win


class TestConf(BackendTest):
    # check_names = False
    # returned_timestamp_unit = 'ns'
    supports_structs = False
    supports_json = False
    supports_arrays = True
    supports_tpch = True
    supports_tpcds = True
    stateful = False
    deps = ("datafusion",)
    # Query 1 seems to require a bit more room here
    tpc_absolute_tolerance = 0.11

    def _load_data(self, **_: Any) -> None:
        con = self.connection
        for table_name in TEST_TABLES:
            path = self.data_dir / "parquet" / f"{table_name}.parquet"
            con.read_parquet(path, table_name=table_name)
        con.create_table("array_types", array_types)
        con.create_table("win", win)
        con.create_table("topk", topk)

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.datafusion.connect(**kw)

    def _load_tpc(self, *, suite, scale_factor):
        con = self.connection
        schema = f"tpc{suite}"
        con.create_database(schema)
        for path in self.data_dir.joinpath(
            schema, f"sf={scale_factor}", "parquet"
        ).glob("*.parquet"):
            table_name = path.with_suffix("").name
            con.con.sql(
                # datafusion can't create an external table in a specific schema it seems
                # so hack around that by
                #
                # 1. creating an external table in the current schema
                # 2. create an internal table in the desired schema using a
                #    CTAS from the external table
                # 3. drop the external table
                f"CREATE EXTERNAL TABLE {table_name} STORED AS PARQUET LOCATION '{path}'"
            )

            con.con.sql(
                f"CREATE TABLE {schema}.{table_name} AS SELECT * FROM {table_name}"
            )
            con.con.sql(f"DROP TABLE {table_name}")

    def _transform_tpc_sql(self, parsed, *, suite, leaves):
        def add_catalog_and_schema(node):
            if isinstance(node, sg.exp.Table) and node.name in leaves:
                return node.__class__(
                    catalog=f"tpc{suite}",
                    **{k: v for k, v in node.args.items() if k != "catalog"},
                )
            return node

        return parsed.transform(add_catalog_and_schema)


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()



from __future__ import annotations

import pytest
from pytest import param


def test_sum(simple_table, assert_sql):
    expr = simple_table.a.sum()
    assert_sql(expr)


def test_count_star(simple_table, assert_sql):
    expr = simple_table.group_by(simple_table.i).size()
    assert_sql(expr)


@pytest.mark.parametrize(
    "unit",
    [
        param("ms", id="timestamp_ms"),
        param("s", id="timestamp_s"),
    ],
)
def test_timestamp_from_unix(simple_table, unit, assert_sql):
    expr = simple_table.d.as_timestamp(unit=unit)
    assert_sql(expr)


def test_complex_projections(simple_table, assert_sql):
    expr = (
        simple_table.group_by(["a", "c"])
        .aggregate(the_sum=simple_table.b.sum())
        .group_by("a")
        .aggregate(mad=lambda x: x.the_sum.abs().mean())
    )
    assert_sql(expr)


def test_filter(simple_table, assert_sql):
    expr = simple_table.filter(
        ((simple_table.c > 0) | (simple_table.c < 0)) & simple_table.g.isin(["A", "B"])
    )
    assert_sql(expr)


@pytest.mark.parametrize(
    "kind",
    [
        "year",
        "quarter",
        "month",
        "week_of_year",
        "day_of_year",
        "day",
        "hour",
        "minute",
        "second",
    ],
)
def test_extract_fields(simple_table, kind, assert_sql):
    expr = getattr(simple_table.i, kind)().name("tmp")
    assert_sql(expr)


def test_complex_groupby_aggregation(simple_table, assert_sql):
    keys = [simple_table.i.year().name("year"), simple_table.i.month().name("month")]
    b_unique = simple_table.b.nunique()
    expr = simple_table.group_by(keys).aggregate(
        total=simple_table.count(), b_unique=b_unique
    )
    assert_sql(expr)


def test_simple_filtered_agg(simple_table, assert_sql):
    expr = simple_table.b.nunique(where=simple_table.g == "A")
    assert_sql(expr)


def test_complex_filtered_agg(simple_table, assert_sql):
    expr = simple_table.group_by("b").aggregate(
        total=simple_table.count(),
        avg_a=simple_table.a.mean(),
        avg_a_A=simple_table.a.mean(where=simple_table.g == "A"),
        avg_a_B=simple_table.a.mean(where=simple_table.g == "B"),
    )
    assert_sql(expr)


def test_value_counts(simple_table, assert_sql):
    expr = simple_table.i.year().value_counts()
    assert_sql(expr)


def test_having(simple_table, assert_sql):
    expr = (
        simple_table.group_by("g")
        .having(simple_table.count() >= 1000)
        .aggregate(simple_table.b.sum().name("b_sum"))
    )
    assert_sql(expr)



from __future__ import annotations

from collections import Counter

import pytest
from pytest import param

import ibis


def test_ifelse_select(backend, alltypes, df):
    table = alltypes
    table = table.select(
        [
            "int_col",
            (
                ibis.ifelse(table["int_col"] == 0, 42, -1)
                .cast("int64")
                .name("where_col")
            ),
        ]
    )

    result = table.execute()

    expected = df.loc[:, ["int_col"]].copy()

    expected["where_col"] = -1
    expected.loc[expected["int_col"] == 0, "where_col"] = 42

    backend.assert_frame_equal(result, expected)


def test_ifelse_column(backend, alltypes, df):
    np = pytest.importorskip("numpy")
    pd = pytest.importorskip("pandas")

    expr = ibis.ifelse(alltypes["int_col"] == 0, 42, -1).cast("int64").name("where_col")
    result = expr.execute()

    expected = pd.Series(
        np.where(df.int_col == 0, 42, -1),
        name="where_col",
        dtype="int64",
    )

    backend.assert_series_equal(result, expected)


def test_substitute(backend):
    val = "400"
    t = backend.functional_alltypes
    expr = (
        t.string_col.nullif("1")
        .substitute({None: val})
        .name("subs")
        .value_counts()
        .filter(lambda t: t.subs == val)
    )
    assert expr["subs_count"].execute()[0] == t.count().execute() // 10


@pytest.mark.parametrize(
    "inp, exp",
    [
        param(
            lambda: ibis.literal(1).cases((1, "one"), (2, "two"), else_="other"),
            "one",
            id="one_kwarg",
        ),
        param(
            lambda: ibis.literal(5).cases((1, "one"), (2, "two")),
            None,
            id="fallthrough",
        ),
    ],
)
def test_value_cases_scalar(con, inp, exp):
    pd = pytest.importorskip("pandas")

    result = con.execute(inp())

    if exp is None:
        assert pd.isna(result)
    else:
        assert result == exp


def test_value_cases_column(batting):
    np = pytest.importorskip("numpy")

    df = batting.to_pandas()
    expr = batting.RBI.cases(
        (5, "five"), (4, "four"), (3, "three"), else_="could be good?"
    )
    result = expr.execute()
    expected = np.select(
        [df.RBI == 5, df.RBI == 4, df.RBI == 3],
        ["five", "four", "three"],
        "could be good?",
    )

    assert Counter(result) == Counter(expected)


def test_ibis_cases_scalar():
    expr = ibis.literal(5).cases((5, "five"), (4, "four"))
    result = expr.execute()
    assert result == "five"


@pytest.mark.notimpl(
    ["sqlite"],
    reason="the int64 RBI column is .to_pandas()ed to an object column, which is incomparable to 5",
    raises=TypeError,
)
def test_ibis_cases_column(batting):
    np = pytest.importorskip("numpy")

    t = batting
    df = batting.to_pandas()
    expr = ibis.cases(
        (t.RBI < 5, "really bad team"), (t.teamID == "PH1", "ph1 team"), else_=t.teamID
    )
    result = expr.execute()
    expected = np.select(
        [df.RBI < 5, df.teamID == "PH1"],
        ["really bad team", "ph1 team"],
        df.teamID,
    )

    assert Counter(result) == Counter(expected)


@pytest.mark.notimpl("clickhouse", reason="special case this and returns 'oops'")
def test_value_cases_null(con):
    """CASE x WHEN NULL never gets hit"""
    e = ibis.literal(5).nullif(5).cases((None, "oops"), else_="expected")
    assert con.execute(e) == "expected"


@pytest.mark.parametrize(
    ("example", "expected"),
    [
        param(lambda: ibis.case().when(True, "yes").end(), "yes", id="top-level-true"),
        param(lambda: ibis.case().when(False, "yes").end(), None, id="top-level-false"),
        param(
            lambda: ibis.case().when(False, "yes").else_("no").end(),
            "no",
            id="top-level-false-value",
        ),
        param(
            lambda: ibis.literal("a").case().when("a", "yes").end(),
            "yes",
            id="method-true",
        ),
        param(
            lambda: ibis.literal("a").case().when("b", "yes").end(),
            None,
            id="method-false",
        ),
        param(
            lambda: ibis.literal("a").case().when("b", "yes").else_("no").end(),
            "no",
            id="method-false-value",
        ),
    ],
)
def test_ibis_case_still_works(con, example, expected):
    # test that the soft-deprecated .case() method still works
    # https://github.com/ibis-project/ibis/pull/9096
    pd = pytest.importorskip("pandas")

    with pytest.warns(FutureWarning):
        expr = example()

    result = con.execute(expr)

    assert (expected is None and pd.isna(result)) or result == expected



from __future__ import annotations

import concurrent.futures
import contextlib
import os
import subprocess
from typing import TYPE_CHECKING, Any

import pytest
import sqlglot as sg

import ibis
import ibis.expr.types as ir
from ibis import util
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping
    from pathlib import Path

CLICKHOUSE_HOST = os.environ.get("IBIS_TEST_CLICKHOUSE_HOST", "localhost")
CLICKHOUSE_PORT = int(os.environ.get("IBIS_TEST_CLICKHOUSE_PORT", 8123))
CLICKHOUSE_USER = os.environ.get("IBIS_TEST_CLICKHOUSE_USER", "default")
CLICKHOUSE_PASS = os.environ.get("IBIS_TEST_CLICKHOUSE_PASSWORD", "")
IBIS_TEST_CLICKHOUSE_DB = os.environ.get("IBIS_TEST_DATA_DB", "ibis_testing")


class TestConf(ServiceBackendTest):
    check_dtype = False
    returned_timestamp_unit = "s"
    supports_json = False
    force_sort = True
    rounding_method = "half_to_even"
    data_volume = "/var/lib/clickhouse/user_files/ibis"
    service_name = "clickhouse"
    deps = ("clickhouse_connect",)
    supports_tpch = True
    supports_tpcds = True
    # Query 14 seems to require a bit more room here
    tpc_absolute_tolerance = 0.0001

    @property
    def native_bool(self) -> bool:
        [(value,)] = self.connection.con.query("SELECT true").result_set
        return isinstance(value, bool)

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("parquet").glob("*.parquet")

    def _load_data(
        self,
        *,
        database: str = IBIS_TEST_CLICKHOUSE_DB,
        **_,
    ) -> None:
        """Load test data into a ClickHouse backend instance.

        Parameters
        ----------
        data_dir
            Location of test data
        script_dir
            Location of scripts defining schemas
        """
        import clickhouse_connect as cc

        con = self.connection
        client = con.con

        with contextlib.suppress(cc.driver.exceptions.DatabaseError):
            client.command(f"CREATE DATABASE {database} ENGINE = Atomic")

        util.consume(map(client.command, self.ddl_script))

    def postload(self, **kw: Any):
        # reconnect to set the database to the test database
        self.connection = self.connect(database=IBIS_TEST_CLICKHOUSE_DB, **kw)

    @staticmethod
    def connect(
        *, tmpdir, worker_id, settings: Mapping[str, Any] | None = None, **kw: Any
    ):
        if settings is None:
            settings = {}

        # without this setting TPC-DS 19 and 24 will fail
        settings.setdefault("allow_experimental_join_condition", 1)

        return ibis.clickhouse.connect(
            host=CLICKHOUSE_HOST,
            port=CLICKHOUSE_PORT,
            password=CLICKHOUSE_PASS,
            user=CLICKHOUSE_USER,
            settings=settings,
            **kw,
        )

    @staticmethod
    def greatest(f: Callable[..., ir.Value], *args: ir.Value) -> ir.Value:
        if len(args) > 2:
            raise NotImplementedError(
                "Clickhouse does not support more than 2 arguments to greatest"
            )
        return f(*args)

    @staticmethod
    def least(f: Callable[..., ir.Value], *args: ir.Value) -> ir.Value:
        if len(args) > 2:
            raise NotImplementedError(
                "Clickhouse does not support more than 2 arguments to least"
            )
        return f(*args)

    def preload(self):
        super().preload()

        suites = ("tpch", "tpcds")

        service_name = self.service_name
        data_volume = self.data_volume

        for suite in suites:
            subprocess.run(
                [
                    "docker",
                    "compose",
                    "exec",
                    service_name,
                    "mkdir",
                    "-p",
                    f"{data_volume}/{suite}",
                ],
                check=True,
            )

        with concurrent.futures.ThreadPoolExecutor() as executor:
            for fut in concurrent.futures.as_completed(
                executor.submit(
                    subprocess.run,
                    [
                        "docker",
                        "compose",
                        "cp",
                        str(path),
                        f"{service_name}:{data_volume}/{suite}/{path.name}",
                    ],
                    check=True,
                )
                for suite in suites
                for path in self.data_dir.joinpath(suite).rglob("*.parquet")
            ):
                fut.result()

    def _load_tpc(self, *, suite, scale_factor):
        con = self.connection
        schema = f"tpc{suite}"
        con.con.command(f"CREATE DATABASE IF NOT EXISTS {schema}")
        parquet_dir = self.data_dir.joinpath(schema, f"sf={scale_factor}", "parquet")
        assert parquet_dir.exists(), parquet_dir
        for path in parquet_dir.glob("*.parquet"):
            table_name = path.with_suffix("").name
            con.con.command(
                f"CREATE VIEW IF NOT EXISTS {schema}.{table_name} AS "
                f"SELECT * FROM file('ibis/{schema}/{path.name}', 'Parquet')"
            )

    def _transform_tpc_sql(self, parsed, *, suite, leaves):
        def add_catalog_and_schema(node):
            if isinstance(node, sg.exp.Table) and node.name in leaves:
                return node.__class__(
                    catalog=f"tpc{suite}",
                    **{k: v for k, v in node.args.items() if k != "catalog"},
                )
            return node

        return parsed.transform(add_catalog_and_schema)


@pytest.fixture(scope="session")
def con(tmp_path_factory, data_dir, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def alltypes(con):
    return con.tables.functional_alltypes


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()



from __future__ import annotations

import re
import string

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.legacy.udf.vectorized as udf
from ibis import util
from ibis.common.graph import Node as Traversable
from ibis.expr.format import fmt, pretty


@pytest.mark.parametrize("cls", [ops.PhysicalTable, ops.Relation])
def test_tables_have_format_value_rules(cls):
    assert cls in fmt.registry


def test_format_table_column(alltypes, snapshot):
    # GH #507
    result = repr(alltypes.f)
    assert "float64" in result
    snapshot.assert_match(result, "repr.txt")


def test_format_projection(alltypes, snapshot):
    # This should produce a ref to the projection
    proj = alltypes[["c", "a", "f"]]
    expr = proj["a"]
    result = repr(expr)
    snapshot.assert_match(result, "repr.txt")


def test_format_table_with_empty_schema(snapshot):
    # GH #6837
    schema = ibis.table({}, name="t")
    result = repr(schema)
    snapshot.assert_match(result, "repr.txt")


def test_table_type_output(snapshot):
    foo = ibis.table(
        [
            ("job", "string"),
            ("dept_id", "string"),
            ("year", "int32"),
            ("y", "double"),
        ],
        name="foo",
    )

    expr = foo.dept_id == foo.view().dept_id
    result = repr(expr)
    assert "UnboundTable: foo" in result
    snapshot.assert_match(result, "repr.txt")


def test_aggregate_arg_names(alltypes, snapshot):
    # Not sure how to test this *well*
    t = alltypes

    by_exprs = [t.g.name("key1"), t.f.round().name("key2")]
    metrics = [t.c.sum().name("c"), t.d.mean().name("d")]

    expr = t.group_by(by_exprs).aggregate(metrics)
    result = repr(expr)
    assert "metrics" in result
    assert "groups" in result

    snapshot.assert_match(result, "repr.txt")


def test_format_multiple_join_with_projection(snapshot):
    # Star schema with fact table
    table = ibis.table(
        [
            ("c", "int32"),
            ("f", "double"),
            ("foo_id", "string"),
            ("bar_id", "string"),
        ],
        "one",
    )

    table2 = ibis.table([("foo_id", "string"), ("value1", "double")], "two")

    table3 = ibis.table([("bar_id", "string"), ("value2", "double")], "three")

    filtered = table.filter(table["f"] > 0)

    pred1 = filtered["foo_id"] == table2["foo_id"]
    pred2 = filtered["bar_id"] == table3["bar_id"]

    j1 = filtered.left_join(table2, [pred1])
    j2 = j1.inner_join(table3, [pred2])

    # Project out the desired fields
    view = j2.select(filtered, table2["value1"], table3["value2"])

    # it works!
    result = repr(view)
    snapshot.assert_match(result, "repr.txt")


def test_memoize_filtered_table(snapshot):
    airlines = ibis.table(
        [("dest", "string"), ("origin", "string"), ("arrdelay", "int32")],
        "airlines",
    )

    dests = ["ORD", "JFK", "SFO"]
    t = airlines.filter(airlines.dest.isin(dests))
    delay_filter = t.dest.topk(10, by=t.arrdelay.mean())

    result = repr(delay_filter)
    snapshot.assert_match(result, "repr.txt")


def test_named_value_expr_show_name(alltypes, snapshot):
    expr = alltypes.f * 2
    expr2 = expr.name("baz")

    # it works!
    result = repr(expr)
    result2 = repr(expr2)

    assert "baz" not in result
    assert "baz" in result2

    snapshot.assert_match(result, "repr.txt")
    snapshot.assert_match(result2, "repr2.txt")


def test_memoize_filtered_tables_in_join(snapshot):
    # related: GH #667
    purchases = ibis.table(
        [
            ("region", "string"),
            ("kind", "string"),
            ("user", "int64"),
            ("amount", "double"),
        ],
        "purchases",
    )

    metric = purchases.amount.sum().name("total")
    agged = purchases.group_by(["region", "kind"]).aggregate(metric)

    left = agged.filter(agged.kind == "foo")
    right = agged.filter(agged.kind == "bar")

    cond = left.region == right.region
    joined = left.join(right, cond).select(left, right.total.name("right_total"))

    result = repr(joined)
    snapshot.assert_match(result, "repr.txt")


def test_argument_repr_shows_name(snapshot):
    t = ibis.table([("fakecolname1", "int64")], name="fakename2")
    expr = t.fakecolname1.nullif(2)
    result = repr(expr)

    assert "fakecolname1" in result
    assert "fakename2" in result
    snapshot.assert_match(result, "repr.txt")


def test_scalar_parameter_formatting():
    value = ibis.param("array<date>")
    assert re.match(r"^param_\d+: \$\(array<date>\)$", str(value)) is not None

    value = ibis.param("int64").name("my_param")
    assert str(value) == "my_param: $(int64)"


def test_same_column_multiple_aliases(snapshot):
    table = ibis.table([("col", "int64")], name="t")
    expr = table.select(table.col.name("fakealias1"), table.col.name("fakealias2"))
    result = repr(expr)

    assert "UnboundTable: t" in result
    assert "col int64" in result
    assert "fakealias1: r0.col" in result
    assert "fakealias2: r0.col" in result
    snapshot.assert_match(result, "repr.txt")


def test_scalar_parameter_repr():
    value = ibis.param(dt.timestamp).name("value")
    assert repr(value) == "value: $(timestamp)"


def test_repr_exact(snapshot):
    # NB: This is the only exact repr test. Do
    # not add new exact repr tests. New repr tests
    # should only check for the presence of substrings.
    table = ibis.table(
        [("col", "int64"), ("col2", "string"), ("col3", "double")],
        name="t",
    ).mutate(col4=lambda t: t.col2.length())

    result = repr(table)
    snapshot.assert_match(result, "repr.txt")


def test_complex_repr(snapshot):
    t = (
        ibis.table(dict(a="int64"), name="t")
        .filter([lambda t: t.a < 42, lambda t: t.a >= 42])
        .mutate(x=lambda t: t.a + 42)
        .group_by("x")
        .aggregate(y=lambda t: t.a.sum())
        .limit(10)
    )
    result = repr(t)

    snapshot.assert_match(result, "repr.txt")


def test_value_exprs_repr():
    t = ibis.table(dict(a="int64", b="string"), name="t")
    assert "r0.a" in repr(t.a)
    assert "Sum(r0.a)" in repr(t.a.sum())


def test_show_types(monkeypatch):
    monkeypatch.setattr(ibis.options.repr, "show_types", True)

    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a / 1.0
    assert "# int64" in repr(t.a)
    assert "# float64" in repr(expr)
    assert "# float64" in repr(expr.sum())


def test_schema_truncation(monkeypatch, snapshot):
    schema = dict(zip(string.ascii_lowercase[:20], ["string"] * 20))
    t = ibis.table(schema, name="t")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 0)
    with pytest.raises(ValueError):
        repr(t)

    monkeypatch.setattr(ibis.options.repr, "table_columns", 1)
    result = repr(t)
    assert util.VERTICAL_ELLIPSIS not in result
    snapshot.assert_match(result, "repr1.txt")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 8)
    result = repr(t)
    assert util.VERTICAL_ELLIPSIS in result
    snapshot.assert_match(result, "repr8.txt")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 1000)
    result = repr(t)
    assert util.VERTICAL_ELLIPSIS not in result
    snapshot.assert_match(result, "repr_all.txt")


def test_table_count_expr(snapshot):
    t1 = ibis.table([("a", "int"), ("b", "float")], name="t1")
    t2 = ibis.table([("a", "int"), ("b", "float")], name="t2")

    cnt = t1.count()
    join_cnt = t1.join(t2, t1.a == t2.a).count()
    union_cnt = ibis.union(t1, t2).count()

    snapshot.assert_match(repr(cnt), "cnt_repr.txt")
    snapshot.assert_match(repr(join_cnt), "join_repr.txt")
    snapshot.assert_match(repr(union_cnt), "union_repr.txt")


def test_window_no_group_by(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a.mean().over(ibis.window(preceding=0))
    result = repr(expr)

    assert "group_by=[]" not in result
    snapshot.assert_match(result, "repr.txt")


def test_window_group_by(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a.mean().over(ibis.window(group_by=t.b))

    result = repr(expr)
    assert "start=0" not in result
    assert "group_by=[r0.b]" in result
    snapshot.assert_match(result, "repr.txt")


def test_fill_null(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")

    expr = t.fill_null({"a": 3})
    snapshot.assert_match(repr(expr), "fill_null_dict_repr.txt")

    expr = t[["a"]].fill_null(3)
    snapshot.assert_match(repr(expr), "fill_null_int_repr.txt")

    expr = t[["b"]].fill_null("foo")
    snapshot.assert_match(repr(expr), "fill_null_str_repr.txt")


def test_asof_join(snapshot):
    left = ibis.table([("time1", "int32"), ("value", "double")], name="left")
    right = ibis.table([("time2", "int32"), ("value2", "double")], name="right")
    right_ = right.view()
    joined = left.asof_join(right, ("time1", "time2")).inner_join(
        right_, left.value == right_.value2
    )

    result = repr(joined)
    snapshot.assert_match(result, "repr.txt")


def test_two_inner_joins(snapshot):
    left = ibis.table(
        [("time1", "int32"), ("value", "double"), ("a", "string")], name="left"
    )
    right = ibis.table(
        [("time2", "int32"), ("value2", "double"), ("b", "string")], name="right"
    )
    right_ = right.view()
    joined = left.inner_join(right, left.a == right.b).inner_join(
        right_, left.value == right_.value2
    )

    result = repr(joined)
    snapshot.assert_match(result, "repr.txt")


def test_destruct_selection(snapshot):
    table = ibis.table([("col", "int64")], name="t")

    with pytest.warns(FutureWarning, match="v9.0"):

        @udf.reduction(
            input_type=["int64"],
            output_type=dt.Struct({"sum": "int64", "mean": "float64"}),
        )
        def multi_output_udf(v):
            return v.sum(), v.mean()

    expr = table.aggregate(multi_output_udf(table["col"]).destructure())
    result = repr(expr)

    snapshot.assert_match(result, "repr.txt")


@pytest.mark.parametrize(
    "literal, typ, output",
    [(42, None, "42"), ("42", None, "'42'"), (42, "double", "42.0")],
)
def test_format_literal(literal, typ, output):
    expr = ibis.literal(literal, type=typ)
    assert repr(expr) == output


def test_format_dummy_table(snapshot):
    t = ops.DummyTable({"foo": ibis.array([1]).cast("array<int8>")}).to_expr()

    result = repr(t)
    snapshot.assert_match(result, "repr.txt")


def test_format_in_memory_table(snapshot):
    pytest.importorskip("pandas")

    t = ibis.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.x.sum() + t.y.sum()

    result = repr(expr)
    assert "InMemoryTable" in result
    snapshot.assert_match(result, "repr.txt")


def test_format_unbound_table_namespace(snapshot):
    t = ibis.table(name="bork", schema=(("a", "int"), ("b", "int")))

    result = repr(t)
    snapshot.assert_match(result, "repr.txt")

    t = ibis.table(name="bork", schema=(("a", "int"), ("b", "int")), database="bork")

    result = repr(t)
    snapshot.assert_match(result, "reprdb.txt")

    t = ibis.table(
        name="bork", schema=(("a", "int"), ("b", "int")), catalog="ork", database="bork"
    )

    result = repr(t)
    snapshot.assert_match(result, "reprcatdb.txt")


def test_format_new_relational_operation(alltypes, snapshot):
    class MyRelation(ops.Relation):
        parent: ops.Relation
        kind: str

        @property
        def schema(self):
            return self.parent.schema

        @property
        def values(self):
            return {}

    table = MyRelation(alltypes, kind="foo").to_expr()
    expr = table.select(table, table.a.name("a2"))
    result = repr(expr)

    snapshot.assert_match(result, "repr.txt")


def test_format_new_value_operation(alltypes, snapshot):
    class Inc(ops.Value):
        arg: ops.Value

        @property
        def dtype(self):
            return self.arg.dtype

        @property
        def shape(self):
            return self.arg.shape

    expr = Inc(alltypes.a).to_expr().name("incremented")
    result = repr(expr)
    last_line = result.splitlines()[-1]

    assert "Inc" in result
    assert last_line == "incremented: Inc(r0.a)"


def test_format_show_variables(monkeypatch, alltypes, snapshot):
    monkeypatch.setattr(ibis.options.repr, "show_variables", True)

    filtered = alltypes.filter(alltypes.f > 0)
    ordered = filtered.order_by("f")
    projected = ordered[["a", "b", "f"]]

    add = projected.a + projected.b
    sub = projected.a - projected.b
    expr = add * sub

    result = repr(expr)

    assert "projected.a" in result
    assert "projected.b" in result
    assert "filtered" in result
    assert "ordered" in result

    snapshot.assert_match(result, "repr.txt")


def test_default_format_implementation(snapshot):
    class ValueList(ops.Node):
        values: tuple[ops.Value, ...]

    t = ibis.table([("a", "int64")], name="t")
    vl = ValueList((1, 2.0, "three", t.a))
    result = pretty(vl)

    snapshot.assert_match(result, "repr.txt")


def test_arbitrary_traversables_are_supported(snapshot):
    class MyNode(Traversable):
        __slots__ = ("children", "obj")
        __argnames__ = ("obj", "children")

        def __init__(self, obj, children):
            self.obj = obj.op()
            self.children = tuple(child.op() for child in children)

        @property
        def __args__(self):
            return self.obj, self.children

        def __hash__(self):
            return hash((self.__class__, self.obj, self.children))

    t = ibis.table([("a", "int64")], name="t")
    node = MyNode(t.a, [t.a, t.a + 1])
    result = pretty(node)

    snapshot.assert_match(result, "repr.txt")



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.expr.operations as ops
from ibis import _
from ibis.common.annotations import ValidationError
from ibis.common.deferred import Deferred
from ibis.common.exceptions import IbisTypeError


@pytest.mark.parametrize(
    ("fn", "operation"),
    [
        param(
            lambda t, where: t.int_col.nunique(where=where),
            ops.CountDistinct,
            id="nunique",
        ),
        param(lambda t, where: t.bool_col.any(where=where), ops.Any, id="any"),
        param(lambda t, where: t.bool_col.all(where=where), ops.All, id="all"),
        param(lambda t, where: t.int_col.sum(where=where), ops.Sum, id="sum"),
        param(lambda t, where: t.int_col.mean(where=where), ops.Mean, id="mean"),
        param(lambda t, where: t.int_col.min(where=where), ops.Min, id="min"),
        param(lambda t, where: t.int_col.max(where=where), ops.Max, id="max"),
        param(
            lambda t, where: t.int_col.argmin(t.string_col, where=where),
            ops.ArgMin,
            id="argmin",
        ),
        param(
            lambda t, where: t.int_col.argmax(t.string_col, where=where),
            ops.ArgMax,
            id="argmax",
        ),
        param(
            lambda t, where: t.int_col.std(how="pop", where=where),
            ops.StandardDev,
            id="std",
        ),
        param(lambda t, where: t.int_col.var(where=where), ops.Variance, id="var"),
        param(
            lambda t, where: t.int_col.approx_nunique(where=where),
            ops.ApproxCountDistinct,
            id="approx_nunique",
        ),
        param(
            lambda t, where: t.int_col.arbitrary(where=where),
            ops.Arbitrary,
            id="arbitrary",
        ),
        param(lambda t, where: t.int_col.first(where=where), ops.First, id="first"),
        param(lambda t, where: t.int_col.last(where=where), ops.Last, id="last"),
        param(
            lambda t, where: t.int_col.bit_and(where=where), ops.BitAnd, id="bit_and"
        ),
        param(lambda t, where: t.int_col.bit_or(where=where), ops.BitOr, id="bit_or"),
        param(
            lambda t, where: t.int_col.bit_xor(where=where), ops.BitXor, id="bit_xor"
        ),
        param(
            lambda t, where: t.int_col.collect(where=where),
            ops.ArrayCollect,
            id="collect",
        ),
        param(
            lambda t, where: t.int_col.approx_quantile(0.5, where=where),
            ops.ApproxQuantile,
            id="approx_quantile",
        ),
        param(
            lambda t, where: t.int_col.approx_quantile([0.25, 0.5, 0.75], where=where),
            ops.ApproxMultiQuantile,
            id="approx_multi_quantile",
        ),
    ],
)
@pytest.mark.parametrize(
    "cond",
    [
        pytest.param(lambda t: None, id="no_cond"),
        pytest.param(
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        pytest.param(
            lambda t: _.string_col.isin(["1", "7"]),
            id="is_in_deferred",
        ),
    ],
)
def test_reduction_methods(fn, operation, cond):
    t = ibis.table(
        name="t",
        schema={
            "string_col": "string",
            "int_col": "int64",
            "bool_col": "boolean",
        },
    )
    where = cond(t)
    expr = fn(t, where)
    node = expr.op()
    assert isinstance(node, operation)
    if where is None:
        assert node.where is None
    elif isinstance(where, Deferred):
        resolved = where.resolve(t).op()
        assert node.where == resolved
    else:
        assert node.where == where.op()


@pytest.mark.parametrize("func_name", ["argmin", "argmax"])
def test_argminmax_deferred(func_name):
    t = ibis.table({"a": "int", "b": "int"}, name="t")
    func = getattr(t.a, func_name)
    assert func(_.b).equals(func(t.b))


@pytest.mark.parametrize("func_name", ["cov", "corr"])
def test_cov_corr_deferred(func_name):
    t = ibis.table({"a": "int", "b": "int"}, name="t")
    func = getattr(t.a, func_name)
    assert func(_.b).equals(func(t.b))


@pytest.mark.parametrize("method", ["collect", "first", "last", "group_concat"])
def test_ordered_aggregations(method):
    t = ibis.table({"a": "string", "b": "int", "c": "int"}, name="t")
    func = getattr(t.a, method)

    q1 = func(order_by="b")
    q2 = func(order_by=("b",))
    q3 = func(order_by=_.b)
    q4 = func(order_by=t.b)
    assert q1.equals(q2)
    assert q1.equals(q3)
    assert q1.equals(q4)

    q5 = func(order_by=("b", "c"))
    q6 = func(order_by=(_.b, _.c))
    assert q5.equals(q6)

    q7 = func(order_by=_.b.desc())
    q8 = func(order_by=t.b.desc())
    assert q7.equals(q8)

    with pytest.raises(IbisTypeError):
        func(order_by="oops")


@pytest.mark.parametrize("method", ["collect", "first", "last", "group_concat"])
def test_ordered_aggregations_no_order(method):
    t = ibis.table({"a": "string", "b": "int", "c": "int"}, name="t")
    func = getattr(t.a, method)

    q1 = func()
    q2 = func(order_by=None)
    q3 = func(order_by=())
    assert q1.equals(q2)
    assert q1.equals(q3)


def test_collect_distinct():
    t = ibis.table({"a": "string", "b": "int", "c": "int"}, name="t")
    # Fine
    t.a.collect(distinct=True)
    t.a.collect(distinct=True, order_by=t.a.desc())
    (t.a + 1).collect(distinct=True, order_by=(t.a + 1).desc())

    with pytest.raises(ValidationError, match="only order by the collected column"):
        t.b.collect(distinct=True, order_by=t.a)
    with pytest.raises(ValidationError, match="only order by the collected column"):
        t.b.collect(
            distinct=True,
            order_by=(
                t.a,
                t.b,
            ),
        )



import ibis


employee = ibis.table(
    name="employee",
    schema={"first_name": "string", "last_name": "string", "id": "int64"},
)
call = ibis.table(
    name="call",
    schema={
        "start_time": "timestamp",
        "end_time": "timestamp",
        "employee_id": "int64",
        "call_outcome_id": "int64",
        "call_attempts": "int64",
    },
)
joinchain = employee.right_join(
    call, [(employee.id == call.employee_id), ibis.literal(True)]
).select(
    employee.first_name,
    employee.last_name,
    employee.id,
    call.start_time,
    call.end_time,
    call.employee_id,
    call.call_outcome_id,
    call.call_attempts,
)
f = joinchain.filter((joinchain.id < 5))
s = f.order_by(f.id.desc())

result = s.select(
    s.first_name,
    s.last_name,
    s.id,
    s.start_time,
    s.end_time,
    s.employee_id,
    s.call_outcome_id,
    s.call_attempts,
    s.first_name.name("first"),
)



from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.sql.compilers.pyspark import PySparkCompiler
from ibis.backends.sql.dialects import Databricks


class DatabricksCompiler(PySparkCompiler):
    __slots__ = ()
    dialect = Databricks

    SIMPLE_OPS = PySparkCompiler.SIMPLE_OPS | {
        ops.Divide: "try_divide",
        ops.Mode: "mode",
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.TypeOf: "typeof",
        ops.RandomUUID: "uuid",
        ops.StringSplit: "split",
    }

    UNSUPPORTED_OPS = (
        ops.ElementWiseVectorizedUDF,
        ops.AnalyticVectorizedUDF,
        ops.ReductionVectorizedUDF,
        ops.RowID,
        ops.TimestampBucket,
    )

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_binary():
            return self.f.unhex(value.hex())
        elif dtype.is_decimal():
            if value.is_finite():
                return self.cast(str(value), dtype)
            else:
                return self.cast(str(value), dt.float64)
        elif dtype.is_uuid():
            return sge.convert(str(value))
        else:
            return None

    def visit_Field(self, op, *, rel, name):
        return sg.column(
            self._gen_valid_name(name), table=rel.alias_or_name, quoted=self.quoted
        )


compiler = DatabricksCompiler()
